from datetime import datetime
from optparse import make_option
from django.core.management.base import NoArgsCommand, CommandError
from django.db.models import Sum
from django.utils.timezone import now, get_current_timezone
from oscar_docdata.models import DocdataOrder


class Command(NoArgsCommand):
    help = "Show the monthly order statistics"
    option_list = (
        make_option('-s', '--status', action='store', dest='status', default='',
            help="Which status to use, defaults to 'paid'"),
    ) + NoArgsCommand.option_list

    def handle_noargs(self, **options):
        """
        Update the status.
        """
        status = options.get('status', DocdataOrder.STATUS_PAID)
        all_status_choices = dict(DocdataOrder.STATUS_CHOICES).keys()
        if status and status not in all_status_choices:
            raise CommandError("Invalid status, valid choices are: {0}".format(", ".join(sorted(all_status_choices))))

        base_qs = DocdataOrder.objects.current_merchant()

        try:
            start_date = base_qs.values_list('created', flat=True).order_by('created')[0]
        except IndexError:
            self.stdout.write("No orders available")

        currencies = list(base_qs.values_list('currency', flat=True).order_by('currency').distinct())

        tzinfo = get_current_timezone()
        cur_date = datetime(start_date.year, start_date.month, 1, tzinfo=tzinfo)
        end_date = now()

        col_style = "| {0:8} | {1:3} | {2:12} | {3:12} | {4:12} | {5:12} | {6:12} |"
        header = col_style.format("Month", "Cur", "Registrered", "Captured", "Difference", "Refunded", "Charged back")
        sep = '-' * len(header)
        self.stdout.write(sep)
        self.stdout.write(header)
        self.stdout.write(sep)

        next_date = None

        while cur_date < end_date:
            if next_date and len(currencies) > 1:
                self.stdout.write(col_style.format('', '', '', '', '', '', ''))

            if cur_date.month == 12:
                next_date = datetime(cur_date.year + 1, 1, 1, tzinfo=tzinfo)
            else:
                next_date = datetime(cur_date.year, cur_date.month + 1, 1, tzinfo=tzinfo)

            qs = base_qs.filter(created__range=(cur_date, next_date))
            if status:
                qs = qs.filter(status=status)

            for currency in currencies:
                totals = qs.filter(currency=currency).aggregate(
                    cap=Sum('total_captured'),
                    reg=Sum('total_registered'),
                    ref=Sum('total_refunded'),
                    cb=Sum('total_charged_back'),
                )

                self.stdout.write(col_style.format(
                    "{0}-{1:02d}".format(cur_date.year, cur_date.month),
                    currency,
                    totals['cap'] or 0,
                    totals['reg'] or 0,
                    (totals['cap'] or 0) - (totals['reg'] or 0) or '',
                    totals['ref'] or 0,
                    totals['cb'] or 0,
                ))

            cur_date = next_date



from django.dispatch import Signal

# The signal which is fired on order status changes.
# You should handle this signal to create the payment Source records in Oscar.
order_status_changed = Signal(providing_args=["order", "old_status", "new_status"])


payment_updated = Signal(providing_args=['order', 'payment'])

payment_added = Signal(providing_args=['order', 'payment'])

return_view_called = Signal(providing_args=['request', 'order', 'callback'])

status_changed_view_called = Signal(providing_args=['request', 'order'])



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'DocdataOrder'
        db.create_table(u'oscar_docdata_docdataorder', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('merchant_order_id', self.gf('django.db.models.fields.CharField')(default='', max_length=100, blank=True)),
            ('order_key', self.gf('django.db.models.fields.CharField')(default='', unique=True, max_length=200, blank=True)),
            ('status', self.gf('django.db.models.fields.CharField')(default='new', max_length=50)),
            ('language', self.gf('django.db.models.fields.CharField')(default='en', max_length=5, blank=True)),
            ('total_gross_amount', self.gf('django.db.models.fields.DecimalField')(max_digits=15, decimal_places=2)),
            ('currency', self.gf('django.db.models.fields.CharField')(max_length=10)),
            ('country', self.gf('django.db.models.fields.CharField')(max_length=2, null=True, blank=True)),
            ('created', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, blank=True)),
            ('updated', self.gf('django.db.models.fields.DateTimeField')(auto_now=True, blank=True)),
        ))
        db.send_create_signal(u'oscar_docdata', ['DocdataOrder'])

        # Adding model 'DocdataPayment'
        db.create_table(u'oscar_docdata_docdatapayment', (
            ('polymorphic_ctype', self.gf('django.db.models.fields.related.ForeignKey')(related_name=u'polymorphic_oscar_docdata.docdatapayment_set', null=True, to=orm['contenttypes.ContentType'])),
            ('docdata_order', self.gf('django.db.models.fields.related.ForeignKey')(related_name='payments', to=orm['oscar_docdata.DocdataOrder'])),
            ('payment_id', self.gf('django.db.models.fields.CharField')(default='', max_length=100, primary_key=True)),
            ('status', self.gf('django.db.models.fields.CharField')(default='NEW', max_length=30)),
            ('payment_method', self.gf('django.db.models.fields.CharField')(default='', max_length=60, blank=True)),
            ('created', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, blank=True)),
            ('updated', self.gf('django.db.models.fields.DateTimeField')(auto_now=True, blank=True)),
        ))
        db.send_create_signal(u'oscar_docdata', ['DocdataPayment'])

        # Adding model 'DocdataDirectDebitPayment'
        db.create_table(u'oscar_docdata_docdatadirectdebitpayment', (
            (u'docdatapayment_ptr', self.gf('django.db.models.fields.related.OneToOneField')(to=orm['oscar_docdata.DocdataPayment'], unique=True, primary_key=True)),
            ('holder_name', self.gf('django.db.models.fields.CharField')(max_length=35)),
            ('holder_city', self.gf('django.db.models.fields.CharField')(max_length=35)),
            ('holder_country_code', self.gf('django.db.models.fields.CharField')(max_length=2, null=True, blank=True)),
            ('iban', self.gf('django.db.models.fields.CharField')(max_length=34)),
            ('bic', self.gf('django.db.models.fields.CharField')(max_length=11)),
        ))
        db.send_create_signal(u'oscar_docdata', ['DocdataDirectDebitPayment'])


    def backwards(self, orm):
        # Deleting model 'DocdataOrder'
        db.delete_table(u'oscar_docdata_docdataorder')

        # Deleting model 'DocdataPayment'
        db.delete_table(u'oscar_docdata_docdatapayment')

        # Deleting model 'DocdataDirectDebitPayment'
        db.delete_table(u'oscar_docdata_docdatadirectdebitpayment')


    models = {
        u'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        u'oscar_docdata.docdatadirectdebitpayment': {
            'Meta': {'ordering': "('-created', '-updated')", 'object_name': 'DocdataDirectDebitPayment', '_ormbases': [u'oscar_docdata.DocdataPayment']},
            'bic': ('django.db.models.fields.CharField', [], {'max_length': '11'}),
            u'docdatapayment_ptr': ('django.db.models.fields.related.OneToOneField', [], {'to': u"orm['oscar_docdata.DocdataPayment']", 'unique': 'True', 'primary_key': 'True'}),
            'holder_city': ('django.db.models.fields.CharField', [], {'max_length': '35'}),
            'holder_country_code': ('django.db.models.fields.CharField', [], {'max_length': '2', 'null': 'True', 'blank': 'True'}),
            'holder_name': ('django.db.models.fields.CharField', [], {'max_length': '35'}),
            'iban': ('django.db.models.fields.CharField', [], {'max_length': '34'})
        },
        u'oscar_docdata.docdataorder': {
            'Meta': {'ordering': "('-created', '-updated')", 'object_name': 'DocdataOrder'},
            'country': ('django.db.models.fields.CharField', [], {'max_length': '2', 'null': 'True', 'blank': 'True'}),
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'currency': ('django.db.models.fields.CharField', [], {'max_length': '10'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'language': ('django.db.models.fields.CharField', [], {'default': "'en'", 'max_length': '5', 'blank': 'True'}),
            'merchant_order_id': ('django.db.models.fields.CharField', [], {'default': "''", 'max_length': '100', 'blank': 'True'}),
            'order_key': ('django.db.models.fields.CharField', [], {'default': "''", 'unique': 'True', 'max_length': '200', 'blank': 'True'}),
            'status': ('django.db.models.fields.CharField', [], {'default': "'new'", 'max_length': '50'}),
            'total_gross_amount': ('django.db.models.fields.DecimalField', [], {'max_digits': '15', 'decimal_places': '2'}),
            'updated': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'})
        },
        u'oscar_docdata.docdatapayment': {
            'Meta': {'ordering': "('-created', '-updated')", 'object_name': 'DocdataPayment'},
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'docdata_order': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'payments'", 'to': u"orm['oscar_docdata.DocdataOrder']"}),
            'payment_id': ('django.db.models.fields.CharField', [], {'default': "''", 'max_length': '100', 'primary_key': 'True'}),
            'payment_method': ('django.db.models.fields.CharField', [], {'default': "''", 'max_length': '60', 'blank': 'True'}),
            'polymorphic_ctype': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "u'polymorphic_oscar_docdata.docdatapayment_set'", 'null': 'True', 'to': u"orm['contenttypes.ContentType']"}),
            'status': ('django.db.models.fields.CharField', [], {'default': "'NEW'", 'max_length': '30'}),
            'updated': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'})
        }
    }

    complete_apps = ['oscar_docdata']


from datetime import timedelta
import logging
from optparse import make_option
from django.core.management.base import NoArgsCommand
from django.db import transaction
from django.utils.timezone import now
from oscar_docdata.facade import get_facade
from oscar_docdata.models import DocdataOrder


class Command(NoArgsCommand):
    help = "Mark old open orders as expired"
    option_list = (
        make_option('-p', '--dry-run', action='store_true', dest='dry-run', default=False,
            help="Only list what will change, don't make the actual changes"),
    ) + NoArgsCommand.option_list

    def handle_noargs(self, **options):
        """
        Update the status.
        """
        is_dry_run = options.get('dry-run', False)
        expire_status_choices = (DocdataOrder.STATUS_NEW, DocdataOrder.STATUS_IN_PROGRESS)

        # At -v2 SOAP requests are outputted.
        verbosity = int(options['verbosity'])
        logging.getLogger('suds.transport').setLevel('INFO' if verbosity < 2 else 'DEBUG')

        qs = DocdataOrder.objects.active_merchants() \
            .filter(status__in=expire_status_choices) \
            .filter(created__lt=(now() - timedelta(days=21)))  # 3 weeks, based on manual testing.

        facade = get_facade()

        if is_dry_run:
            self.stdout.write(u"Expiring orders (DRY-RUN):")
        else:
            self.stdout.write(u"Expiring orders:")

        for order in qs.iterator():
            # Will loop through all one by one, so signals can be properly fired:
            self.stdout.write(u"- {0}\t(created {1:%Y-%m-%d}, still {2})".format(order.merchant_order_id, order.created, order.status))

            # Will update
            old_status = order.status
            order.status = DocdataOrder.STATUS_EXPIRED

            if not is_dry_run:
                with transaction.atomic():
                    # First request the order at docdata, avoid expiring an order which missed an update (very unlikely)
                    facade.update_order(order)
                    if order.status not in expire_status_choices:
                        if order.status == DocdataOrder.STATUS_EXPIRED:
                            self.stdout.write(u"  Updated order {0} via status API, detected expired state".format(order.merchant_order_id))
                        else:
                            self.stderr.write(u"  Skipping order {0}, status changed to: {1}".format(order.merchant_order_id, order.status))
                    else:
                        # More efficient SQL
                        DocdataOrder.objects.filter(id=order.id).update(status=DocdataOrder.STATUS_EXPIRED)

                        try:
                            # Make sure Oscar is updated, and the signal is sent.
                            facade.order_status_changed(order, old_status, order.status)
                        except Exception as e:
                            self.stderr.write(u"Failed to update order {0}: {1}".format(order.id, e))
                            DocdataOrder.objects.filter(id=order.id).update(status=old_status)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from decimal import Decimal
from oscar_docdata import appsettings


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='DocdataOrder',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('merchant_name', models.CharField(default=appsettings.DOCDATA_MERCHANT_NAME, max_length=100, verbose_name='Docdata account')),
                ('merchant_order_id', models.CharField(default=b'', max_length=100, verbose_name='Order ID')),
                ('order_key', models.CharField(default=b'', unique=True, max_length=200, verbose_name='Payment cluster ID')),
                ('status', models.CharField(default=b'new', max_length=50, verbose_name='Status', choices=[(b'new', 'New'), (b'in_progress', 'In Progress'), (b'pending', 'Pending'), (b'paid', 'Paid'), (b'paid_refunded', 'Paid, part refunded'), (b'cancelled', 'Cancelled'), (b'charged_back', 'Charged back'), (b'refunded', 'Refunded'), (b'expired', 'Expired'), (b'unknown', 'Unknown')])),
                ('language', models.CharField(default=b'en', max_length=5, verbose_name='Language', blank=True)),
                ('total_gross_amount', models.DecimalField(verbose_name='Total gross amount', max_digits=15, decimal_places=2)),
                ('currency', models.CharField(max_length=10, verbose_name='Currency')),
                ('country', models.CharField(max_length=2, null=True, verbose_name='Country_code', blank=True)),
                ('total_registered', models.DecimalField(default=Decimal('0.00'), verbose_name='Total registered', max_digits=15, decimal_places=2)),
                ('total_shopper_pending', models.DecimalField(default=Decimal('0.00'), verbose_name='Total shopper pending', max_digits=15, decimal_places=2)),
                ('total_acquirer_pending', models.DecimalField(default=Decimal('0.00'), verbose_name='Total acquirer pending', max_digits=15, decimal_places=2)),
                ('total_acquirer_approved', models.DecimalField(default=Decimal('0.00'), verbose_name='Total acquirer approved', max_digits=15, decimal_places=2)),
                ('total_captured', models.DecimalField(default=Decimal('0.00'), verbose_name='Total captured', max_digits=15, decimal_places=2)),
                ('total_refunded', models.DecimalField(default=Decimal('0.00'), verbose_name='Total refunded', max_digits=15, decimal_places=2)),
                ('total_charged_back', models.DecimalField(default=Decimal('0.00'), verbose_name='Total changed back', max_digits=15, decimal_places=2)),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='created')),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='updated')),
            ],
            options={
                'ordering': ('-created', '-updated'),
                'verbose_name': 'Docdata Order',
                'verbose_name_plural': 'Docdata Orders',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='DocdataPayment',
            fields=[
                ('payment_id', models.CharField(primary_key=True, default=b'', serialize=False, max_length=100, blank=True, verbose_name='Payment id')),
                ('status', models.CharField(default=b'NEW', max_length=30, verbose_name='status')),
                ('payment_method', models.CharField(default=b'', max_length=60, blank=True)),
                ('confidence_level', models.CharField(default=b'', verbose_name='Confidence level', max_length=30, editable=False)),
                ('amount_allocated', models.DecimalField(default=Decimal('0.00'), verbose_name='Amount Allocated', editable=False, max_digits=12, decimal_places=2)),
                ('amount_debited', models.DecimalField(default=Decimal('0.00'), verbose_name='Amount Debited', editable=False, max_digits=12, decimal_places=2)),
                ('amount_refunded', models.DecimalField(default=Decimal('0.00'), verbose_name='Amount Refunded', editable=False, max_digits=12, decimal_places=2)),
                ('amount_chargeback', models.DecimalField(default=Decimal('0.00'), verbose_name='Amount Changed back', editable=False, max_digits=12, decimal_places=2)),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='created')),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='updated')),
            ],
            options={
                'ordering': ('payment_id',),
                'verbose_name': 'Payment',
                'verbose_name_plural': 'Payments',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='DocdataDirectDebitPayment',
            fields=[
                ('docdatapayment_ptr', models.OneToOneField(parent_link=True, auto_created=True, primary_key=True, serialize=False, to='oscar_docdata.DocdataPayment')),
                ('holder_name', models.CharField(max_length=35)),
                ('holder_city', models.CharField(max_length=35)),
                ('holder_country_code', models.CharField(max_length=2, null=True, verbose_name='Country_code', blank=True)),
                ('iban', models.CharField(max_length=34)),
                ('bic', models.CharField(max_length=11)),
            ],
            options={
                'ordering': ('-created', '-updated'),
                'verbose_name': 'Direct Debit Payment',
                'verbose_name_plural': 'Derect Debit Payments',
            },
            bases=('oscar_docdata.docdatapayment',),
        ),
        migrations.AddField(
            model_name='docdatapayment',
            name='docdata_order',
            field=models.ForeignKey(related_name='payments', to='oscar_docdata.DocdataOrder'),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='docdatapayment',
            name='polymorphic_ctype',
            field=models.ForeignKey(related_name='polymorphic_oscar_docdata.docdatapayment_set+', editable=False, to='contenttypes.ContentType', null=True),
            preserve_default=True,
        ),
    ]



from django.conf.urls import url
from oscar.core.application import Application
from . import views


class DocdataOrderDashboardApplication(Application):
    name = None
    default_permissions = ['is_staff', ]

    list_view = views.DocdataOrderListView
    detail_view = views.DocdataOrderDetailView
    update_status_view = views.DocdataOrderUpdateStatusView
    cancel_view = views.DocdataOrderCancelView

    def get_urls(self):
        """
        Get URL patterns defined for flatpage management application.
        """
        urls = [
            url(r'^$', self.list_view.as_view(), name='docdata-order-list'),
            url(r'^detail/(?P<pk>[-\w]+)/$', self.detail_view.as_view(), name='docdata-order-detail'),
            url(r'^update-status/(?P<pk>[-\w]+)/$', self.update_status_view.as_view(), name='docdata-order-update-status'),
            url(r'^cancel/(?P<pk>[-\w]+)/$', self.cancel_view.as_view(), name='docdata-order-cancel'),
        ]
        return self.post_process_urls(urls)


application = DocdataOrderDashboardApplication()



from six import text_type


class OrderKeyMissing(ValueError):
    """
    Missing an order key!
    """

class InvalidMerchant(ValueError):
    """
    Provided order belongs to a different merchant!
    """

class DocdataException(Exception):
    """
    Base class for all exceptions from Docdata
    """
    def __init__(self, code, value):
        super(DocdataException, self).__init__(value)
        self.code = code
        self.value = text_type(value)


class DocdataCreateError(DocdataException):
    """
    There was an error creating the payment.
    """


class DocdataStartError(DocdataException):
    """
    There was an error start the payment..
    """


class DocdataStatusError(DocdataException):
    """
    There was an error requesting the status
    """


class DocdataCancelError(DocdataException):
    """
    There was an error cancelling the order.
    """



"""
Backend calls to docdata.

Gateway module - this module is ignorant of Oscar and could be used in a non-Oscar project.
All Oscar-related functionality should be in the facade.
"""
import logging
from decimal import Decimal as D
from django.core.exceptions import ImproperlyConfigured
from django.utils.text import Truncator
import suds.client
import suds.plugin
from django.core.urlresolvers import reverse
from django.utils.translation import get_language
from suds.sax.element import Element
from oscar_docdata import appsettings, __version__ as oscar_docdata_version
from oscar_docdata.exceptions import DocdataCreateError, DocdataStatusError, DocdataStartError, DocdataCancelError, OrderKeyMissing
from six import text_type
from six.moves.urllib.parse import urlencode
from six.moves.urllib.error import URLError

logger = logging.getLogger(__name__)

__all__ = (
    'DocdataClient',

    'CreateReply',
    'StartReply',
    'StatusReply',

    'Name',
    'Shopper',
    'Destination',
    'Address',
    'Amount',
    'Vat',
    'Invoice',
    'Item',

    'Payment',
    'AmexPayment',
    'MasterCardPayment',
    'DirectDebitPayment',
    'IdealPayment',
    'BankTransferPayment',
    'ElvPayment',
)


# Thanks to https://github.com/onepercentclub/onepercentclub-site for an example implmentation,
# which is BSD licensed, copyright (c) 2013 1%CLUB and Ben Konrath.


# Getting call metadata:
#
# >>> from suds.client import Client
# >>> url = 'https://test.docdatapayments.com/ps/services/paymentservice/1_0?wsdl'
# >>> client = Client(url)
#
# >>> print client
#
# >>> client.factory.create('ns0:name')

CACHED_CLIENT = {}

def get_suds_client(testing_mode=False):
    """
    Create the suds client to connect to docdata.
    """
    if testing_mode:
        url = 'https://test.docdatapayments.com/ps/services/paymentservice/1_2?wsdl'
    else:
        url = 'https://secure.docdatapayments.com/ps/services/paymentservice/1_2?wsdl'
    # Online preview: https://secure.docdatapayments.com/ps/orderapi-1_2.wsdl

    # See if the client is already fetched, if so, reuse that.
    try:
        return CACHED_CLIENT[url]
    except KeyError:
        pass

    try:
        client = suds.client.Client(url, plugins=[DocdataAPIVersionPlugin()])
        # HACK: Fixes serialization of raw Element objects.
        # Otherwise, the Element is appended as <tagname /> in the request.
        # The debug output of 'suds.client' won't show this,
        # but the debug output of 'suds' will.
        client.options.prettyxml = True

        # Cache and return
        CACHED_CLIENT[url] = client
        return client
    except URLError as e:
        logger.error("Could not initialize SUDS SOAP client to connect to Docdata: {0}".format(str(e)))
        raise


class DocdataAPIVersionPlugin(suds.plugin.MessagePlugin):
    """
    This adds the API version number to the body element. This is required for the Docdata soap API.
    """

    def marshalled(self, context):
        body = context.envelope.getChild('Body')
        request = body[0]
        request.set('version', '1.2')


#class DocdataBrokenWSDLPlugin(suds.plugin.DocumentPlugin):
#    def parsed(self, context):
#        """ Called after parsing a WSDL or XSD document. The context contains the url & document root. """
#        # The WSDL for the live payments API incorrectly references the wrong location.
#        if len(context.document.children) == 19 and len(context.document.children[18]) > 0:
#            location_attribute = context.document.children[18].children[0].getChild('address').attributes[0]
#            location_attribute.setValue('https://secure.docdatapayments.com:443/ps/services/paymentservice/1_0')


def log_docdata_error(soap_error, message, *args, **kwargs):
    logger.error(u"{0}: code={1}, error={2}".format(message, soap_error._code, soap_error.value), *args, **kwargs)


class DocdataClient(object):
    """
    API Client for docdata.

    This is a wrapper around the SOAP service methods,
    providing more Python-friendly wrappers.
    """

    # Status values. Besides the regular ones mentioned in the documentation,
    # the WSDL also mentions additional status values for the authorizationStatus enum.
    # The CANCELLED statuses are actually misspelled in the protocol.
    STATUS_NEW = 'NEW'

    STATUS_RISK_CHECK_OK = 'RISK_CHECK_OK'
    STATUS_RISK_CHECK_FAILED = 'RISK_CHECK_FAILED'

    STATUS_STARTED = 'STARTED'
    STATUS_START_ERROR = 'START_ERROR'

    STATUS_AUTHENTICATED = 'AUTHENTICATED'
    STATUS_REDIRECTED_FOR_AUTHENTICATION = 'REDIRECTED_FOR_AUTHENTICATION'
    STATUS_AUTHENTICATION_FAILED = 'AUTHENTICATION_FAILED'
    STATUS_AUTHENTICATION_ERROR = 'AUTHENTICATION_ERROR'

    STATUS_AUTHORIZED = 'AUTHORIZED'
    STATUS_REDIRECTED_FOR_AUTHORIZATION = 'REDIRECTED_FOR_AUTHORIZATION'
    STATUS_AUTHORIZATION_REQUESTED = 'AUTHORIZATION_REQUESTED'
    STATUS_AUTHORIZATION_FAILED = 'AUTHORIZATION_FAILED'
    STATUS_AUTHORIZATION_ERROR = 'AUTHORIZATION_ERROR'

    STATUS_PAID = 'PAID'

    STATUS_CANCELLED = 'CANCELED'  # Typoo in protocol, not mentioned in the docs
    STATUS_CANCEL_FAILED = 'CANCEL_FAILED'
    STATUS_CANCEL_ERROR = 'CANCEL_ERROR'
    STATUS_CANCEL_REQUESTED = 'CANCEL_REQUESTED'

    STATUS_CHARGED_BACK = 'CHARGED-BACK'
    STATUS_CONFIRMED_PAID = 'CONFIRMED_PAID'
    STATUS_CONFIRMED_CHARGEDBACK = 'CONFIRMED_CHARGEDBACK'
    STATUS_CLOSED_SUCCESS = 'CLOSED_SUCCESS'
    STATUS_CLOSED_CANCELLED = 'CLOSED_CANCELED'  # Typoo in protocol

    DOCUMENTED_STATUS_VALUES = (
        STATUS_NEW,
        STATUS_STARTED,
        STATUS_REDIRECTED_FOR_AUTHENTICATION,
        STATUS_AUTHORIZED,
        STATUS_AUTHORIZATION_REQUESTED,
        STATUS_PAID,
        STATUS_CANCELLED,
        STATUS_CHARGED_BACK,
        STATUS_CONFIRMED_PAID,
        STATUS_CONFIRMED_CHARGEDBACK,
        STATUS_CLOSED_SUCCESS,
        STATUS_CLOSED_CANCELLED,
    )

    SEEN_UNDOCUMENTED_STATUS_VALUES = (
        STATUS_AUTHORIZATION_FAILED,
        STATUS_CANCEL_FAILED,
        STATUS_AUTHORIZATION_ERROR,
    )

    # Payment methods for the start operation.
    PAYMENT_METHOD_AMEX = 'AMEX'
    PAYMENT_METHOD_MASTERCARD = 'MASTERCARD'
    PAYMENT_METHOD_VISA = 'VISA'
    PAYMENT_METHOD_DIRECT_DEBIT = 'DIRECT_DEBIT'
    PAYMENT_METHOD_BANK_TRANSFER = 'BANK_TRANSFER'
    PAYMENT_METHOD_ELV = 'ELV'


    def __init__(self, testing_mode=None, merchant_name=None, merchant_password=None):
        """
        Initialize the client.

        Optionally, custom merchant credentials can be given.
        By default, the settings ``DOCDATA_MERCHANT_NAME`` and ``DOCDATA_MERCHANT_PASSWORD`` will be used.
        """
        if testing_mode is None:
            testing_mode = appsettings.DOCDATA_TESTING

        self.testing_mode = testing_mode
        self.client = get_suds_client(testing_mode)

        # Create the merchant node to pass the username/password to.
        if merchant_name is not None:
            # Received custom object values
            self.set_merchant(merchant_name, merchant_password)
        else:
            # Using default configuration values
            if not appsettings.DOCDATA_MERCHANT_NAME:
                raise ImproperlyConfigured("Missing DOCDATA_MERCHANT_NAME setting!")
            if not appsettings.DOCDATA_MERCHANT_PASSWORD:
                raise ImproperlyConfigured("Missing DOCDATA_MERCHANT_PASSWORD setting!")

            self.set_merchant(appsettings.DOCDATA_MERCHANT_NAME, appsettings.DOCDATA_MERCHANT_PASSWORD)

        # Create the integration info node which is passed to every request.
        self.integration_info = TechnicalIntegrationInfo()

    @classmethod
    def for_merchant(cls, merchant_name, testing_mode=None):
        """
        Generate the client with the proper credentials that is needed to update the order.
        This method is useful when there are multiple sub accounts in use.
        The proper account credentials are automatically selected
        from the ``DOCDATA_MERCHANT_PASSWORDS`` setting.
        :rtype: DocdataClient
        """
        try:
            password = appsettings.DOCDATA_MERCHANT_PASSWORDS[merchant_name]
        except KeyError:
            raise ImproperlyConfigured("No password provided in DOCDATA_MERCHANT_PASSWORDS for merchant '{0}'".format(merchant_name))

        return cls(
            testing_mode=testing_mode,
            merchant_name=merchant_name,
            merchant_password=password
        )

    def set_merchant(self, name, password):
        """
        Set the merchant name and password to connect to Docdata.
        This defines the account used to connect.
        """
        # Create the merchant node which is passed to every request.
        # The _ notation is used to assign attributes to the XML node, instead of child elements.
        self.merchant = self.client.factory.create('ns0:merchant')
        self.merchant._name = name
        self.merchant._password = password

    @property
    def merchant_name(self):
        """
        Access the merchant name used to connect
        :rtype: str
        """
        return self.merchant._name

    def create(self,
            order_id,
            total_gross_amount,
            shopper,
            bill_to,
            description,
            invoice=None,
            receiptText=None,
            includeCosts=False,
            profile=appsettings.DOCDATA_PROFILE,
            days_to_pay=appsettings.DOCDATA_DAYS_TO_PAY,
        ):
        """
        Create the payment in docdata.

        This is the first step of any payment session.
        After the payment is created, an ``order_key`` will be used.
        This key can be used to continue using the Payment Menu,
        or make the next call to start a Web Direct payment.

        The goal of the create operation is solely to create a payment order on Docdata Payments system.
        Creating a payment order is always the first step of any workflow in Docdata Payments payment service.

        After an order is created, payments can be made on this order; either through (the shopper via) the web menu
        or through the API by the merchant. If the order has been created using information on specific order items,
        the web menu can make use of this information by displaying a shopping cart.

        :param order_id: Unique merchant reference to this order.
        :type total_gross_amount: Amount
        :param shopper: Information concerning the shopper who placed the order.
        :type shopper: Shopper
        :param bill_to: Name and address to use for billing.
        :type bill_to: Destination
        :param description: The description of the order (max 50 chars).
        :type description: str
        :param receiptText: The description that is used by payment providers on shopper statements (max 50 chars).
        :type receiptText: str
        :param profile: The profile that is used to select the payment methods that can be used to pay this order.
        :param days_to_pay: The expected number of days in which the payment should be processed, or be expired if not paid.
        :rtype: CreateReply
        """
        # Preferences for the DocData system.
        paymentPreferences = self.client.factory.create('ns0:paymentPreferences')
        paymentPreferences.profile = profile
        paymentPreferences.numberOfDaysToPay = days_to_pay
        # paymentPreferences.exhortation.period1 ?
        # paymentPreferences.exhortation.period2 ?

        # Menu preferences are empty. They are used for CSS file selection in the payment menu.
        menuPreferences = self.client.factory.create('ns0:menuPreferences')

        # Execute create payment order request.
        #
        # create(
        #     merchant merchant, string35 merchantOrderReference, paymentPreferences paymentPreferences,
        #     menuPreferences menuPreferences, shopper shopper, amount totalGrossAmount,
        #     destination billTo, string50 description, string50 receiptText, xs:boolean includeCosts,
        #     paymentRequest paymentRequest, invoice invoice, technicalIntegrationInfo integrationInfo )
        #
        # The WSDL and XSD also contain documentation individual parameters:
        # https://secure.docdatapayments.com/ps/services/paymentservice/1_2?xsd=1
        #
        # This displays the results in the docdata web menu.
        #
        factory = self.client.factory
        reply = self.client.service.create(
            merchant=self.merchant,
            merchantOrderReference=order_id,
            paymentPreferences=paymentPreferences,
            menuPreferences=menuPreferences,
            shopper=shopper.to_xml(factory),
            totalGrossAmount=total_gross_amount.to_xml(factory),
            billTo=bill_to.to_xml(factory),
            description=description or None,
            receiptText=receiptText or None,
            includeCosts=includeCosts or False,
            # paymentRequest=None,
            invoice=invoice.to_xml(factory) if invoice is not None else None,
            integrationInfo=self.integration_info.to_xml(factory)
        )

        # Parse the reply
        if hasattr(reply, 'createSuccess'):
            order_key = str(reply['createSuccess']['key'])
            return CreateReply(order_id, order_key)
        elif hasattr(reply, 'createError'):
            error = reply.createError.error
            log_docdata_error(error, "DocdataClient: failed to create payment for order %s", order_id)
            raise DocdataCreateError(error._code, error.value)
        else:
            raise NotImplementedError('Received unknown reply from DocData. Remote Payment not created.')


    def start(self, order_key, payment, payment_method=None, amount=None):
        """
        The start operation is used for starting a (web direct) payment on an order.
        It does not need to be used if the merchant makes use of Docdata Payments web menu.

        The web direct can be used for recurring payments for example.
        Standard payments (e.g. iDEAL, creditcard) all happen through the web menu
        because implementing those locally requires certification by the credit card companies.

        TODO: untested

        :type order_key: str
        :param payment: A subclass of the payment class, which one depends on the payment method.
        :type payment: Payment
        :param payment_method: One of the supported payment methods, e.g. PAYMENT_METHOD_IDEAL, PAYMENT_METHOD_MASTERCARD.
                               If omitted, the payment method of the ``payment`` object is used.
        :type payment_method: str
        :param amount: Optional payment amount. If left empty, the full amount of the payment order is used.
        :type amount: Amount
        """
        if not order_key:
            raise OrderKeyMissing("Missing order_key!")

        # We only need to set amount because of bug in suds library. Otherwise it defaults to order amount.

        paymentRequestInput = self.client.factory.create('ns0:paymentRequestInput')
        if amount is not None:
            paymentRequestInput.paymentAmount = amount.to_xml(self.client.factory)
        paymentRequestInput.paymentMethod = payment_method or payment.payment_method
        paymentRequestInput[payment.request_parameter] = payment.to_xml(self.client.factory)

        # Execute start payment request.
        reply = self.client.service.start(
            self.merchant,
            order_key,
            paymentRequestInput,
            integrationInfo=self.integration_info.to_xml(self.client.factory)
        )
        if hasattr(reply, 'startSuccess'):
            return StartReply(reply.startSuccess.paymentId)
        elif hasattr(reply, 'startError'):
            error = reply.createError.error
            log_docdata_error(error, "DocdataClient: failed to get start payment for order %s", order_key)
            raise DocdataStartError(error._code, error.value)
        else:
            raise NotImplementedError('Received unknown reply from DocData. Remote Payment not created.')


    def cancel(self, order_key):
        """
        The cancel command is used for canceling a previously created payment,
        and can only be used for payments with status NEW, STARTED and AUTHORIZED.
        """
        if not order_key:
            raise OrderKeyMissing("Missing order_key!")

        reply = self.client.service.cancel(self.merchant, order_key)

        if hasattr(reply, 'cancelSuccess'):
            return True
        elif hasattr(reply, 'cancelError'):
            error = reply.cancelError.error
            log_docdata_error(error, "DocdataClient: failed to cancel the order %s", order_key)
            raise DocdataCancelError(error._code, error.value)
        else:
            logger.error("Unexpected response node from docdata!")
            raise NotImplementedError('Received unknown reply from DocData. Remote Payment not cancelled.')


    def status(self, order_key):
        """
        Request the status of of order and it's payments.

        :rtype: StatusReply
        """
        # Example response:
        #
        # <?xml version='1.0' encoding='UTF-8'?>
        # <statusResponse xmlns="http://www.docdatapayments.com/services/paymentservice/1_0/">
        #   <statusSuccess>
        #     <success code="SUCCESS">Operation successful.</success>
        #     <report>
        #       <approximateTotals exchangedTo="EUR" exchangeRateDate="2012-12-04 14:39:53">
        #         <totalRegistered>3310</totalRegistered>
        #         <totalShopperPending>0</totalShopperPending>
        #         <totalAcquirerPending>0</totalAcquirerPending>
        #         <totalAcquirerApproved>3310</totalAcquirerApproved>
        #         <totalCaptured>0</totalCaptured>
        #         <totalRefunded>0</totalRefunded>
        #         <totalChargedback>0</totalChargedback>
        #       </approximateTotals>
        #       <payment>           # Can occur multiple times.
        #         <id>1606709142</id>
        #         <paymentMethod>MASTERCARD</paymentMethod>
        #         <authorization>
        #           <status>AUTHORIZED</status>
        #           <amount currency="EUR">3310</amount>
        #           <confidenceLevel>ACQUIRER_APPROVED</confidenceLevel>
        #         </authorization>
        #       </payment>
        #     </report>
        #   </statusSuccess>
        # </statusResponse>
        if not order_key:
            raise OrderKeyMissing("Missing order_key!")

        reply = self.client.service.status(
            self.merchant,
            order_key,
            iIntegrationInfo=self.integration_info.to_xml(self.client.factory)  # NOTE: called iIntegrationInfo in the XSD!!
        )

        if hasattr(reply, 'statusSuccess'):
            return StatusReply(order_key, reply.statusSuccess.report)
        elif hasattr(reply, 'statusError'):
            error = reply.statusError.error
            log_docdata_error(error, "DocdataClient: failed to get status for payment cluster %s", order_key)
            raise DocdataStatusError(error._code, error.value)
        else:
            logger.error("Unexpected response node from docdata!")
            raise NotImplementedError('Received unknown reply from DocData. No status processed from Docdata.')


    def status_extended(self, order_key):
        """
        Request the status with extended information.
        """
        if not order_key:
            raise OrderKeyMissing("Missing order_key!")

        reply = self.client.service.statusExtended(
            self.merchant,
            order_key,
            self.integration_info.to_xml(self.client.factory)  # NOTE: called iIntegrationInfo in the XSD!!
        )

        if hasattr(reply, 'statusSuccess'):
            return StatusReply(order_key, reply.statusSuccess.report)
        elif hasattr(reply, 'statusError'):
            error = reply.statusError.error
            log_docdata_error(error, "DocdataClient: failed to get status for payment cluster %s", order_key)
            raise DocdataStatusError(error._code, error.value)
        else:
            logger.error("Unexpected response node from docdata!")
            raise NotImplementedError('Received unknown reply from DocData. Remote Payment not created.')


    def get_payment_menu_url(self, request, order_key, return_url=None, client_language=None, **extra_url_args):
        """
        Return the URL to the payment menu,
        where the user can be redirected to after creating a successful payment.

        Make sure URLs are absolute, and include the hostname and ``https://`` part.

        When using default_act (possible values "yes" or "true") as well as default_pm,
        your shopper will be redirected straight from your shop to the payment page of the payment method.
        This works only with those payment methods that are initiated by a single click,
        PayPal, Rabo SMS-betalen, SofortUberweisung, eBanking, KBC Betaalknop and iDEAL.
        When the issuer_id is added to the redirect string, this works for iDEAL as well.

        :param extra_args: Additional URL arguments, e.g:
                           default_pm=IDEAL
                           ideal_issuer_id=0021
                           default_act='true'
                           merchant_name='account'
        """
        if not return_url:
            return_url = reverse('return_url', current_app='oscar_docdata')

        # Add order_id= parameter to the URL
        if '?' in return_url:
            url_format = '{0}&callback={callback}&order_id={order_id}'
        else:
            url_format = '{0}?callback={callback}&order_id={order_id}'

        # Make sure URLs are absolute.
        if not '://' in return_url or return_url.startswith('/'):
            return_url = request.build_absolute_uri(return_url)

        args = {
            'command': 'show_payment_cluster',
            'payment_cluster_key': order_key,
            'merchant_name': self.merchant_name,
            'return_url_success': url_format.format(return_url, callback='SUCCESS', order_id=order_key),
            'return_url_pending': url_format.format(return_url, callback='PENDING', order_id=order_key),
            'return_url_canceled': url_format.format(return_url, callback='CANCELLED', order_id=order_key),
            'return_url_error': url_format.format(return_url, callback='ERROR', order_id=order_key),
            'client_language': (client_language or get_language()).upper()
        }
        args.update(extra_url_args)

        if self.testing_mode:
            return 'https://test.docdatapayments.com/ps/menu?' + urlencode(args)
        else:
            return 'https://secure.docdatapayments.com/ps/menu?' + urlencode(args)



class CreateReply(object):
    """
    Docdata response for the create request
    """
    def __init__(self, order_id, order_key):
        # In this library, we favor explicit reply objects over dictionaries,
        # because it makes it much more explicit what is being returned.
        # BTW, PyPy also loves this ;) Much easier to optimize then dict lookups.
        self.order_id = order_id
        self.order_key = order_key

    def __repr__(self):
        return "<CreateReply {0}>".format(self.order_key)


class StartReply(object):
    """
    Docdata response for the start request.
    """
    def __init__(self, payment_id):
        self.payment_id = payment_id

    def __repr__(self):
        return "<StartReply {0}>".format(self.payment_id)


class StatusReply(object):
    """
    Docdata response for the status request.
    """
    def __init__(self, order_key, report):
        self.order_key = order_key
        self.report = report

    def __repr__(self):
        return "<StatusReply {0}>".format(repr(self.report))


class Name(object):
    """
    A name for Docdata.

    :type first: str
    :type last: str
    :type middle: str
    :type initials: str
    :type prefix: str
    :type suffix: str
    """
    def __init__(self, first, last, middle=None, initials=None, prefix=None, suffix=None):
        if not last:
            raise ValueError("Name.last is required!")
        if not first:
            raise ValueError("Name.first is required!")
        self.first = first
        self.last = last
        self.prefix = prefix
        self.initials = initials
        self.middle = middle
        self.suffix = suffix

    def to_xml(self, factory):
        # Assigning values is perhaps very Java-esque, but it's very obvious too
        # what's happening here, while keeping Python-like constructor argument styles.
        node = factory.create('ns0:name')
        node.first = text_type(self.first)
        node.middle = text_type(self.middle) if self.middle else None
        node.last = text_type(self.last)
        node.initials = text_type(self.initials) if self.initials else None
        node.prefix = text_type(self.prefix) if self.prefix else None
        node.suffix = text_type(self.suffix) if self.suffix else None
        return node


class Shopper(object):
    """
    Information concerning the shopper who placed the order.

    :type id: long
    :type name: Name
    :type email: str
    :type language: str
    :type gender: str
    :type date_of_birth: :class:`datetime.Date`
    :type phone_number: str
    :type mobile_phone_number: str
    """
    def __init__(self, id, name, email, language, gender="U", date_of_birth=None, phone_number=None, mobile_phone_number=None, ipAddress=None):
        """
        :type name: Name
        """
        self.id = id
        self.name = name
        self.email = email
        self.language = language
        self.gender = gender   # M (male), F (female), U (undefined)
        self.date_of_birth = date_of_birth
        self.phone_number = phone_number
        self.mobile_phone_number = mobile_phone_number  # +316..
        self.ipAddress = ipAddress

    def to_xml(self, factory):
        language_node = factory.create('ns0:language')
        language_node._code = self.language

        node = factory.create('ns0:shopper')
        if self.id is not None and self.id != '':
            node._id = self.id  # attribute, hence the ._id
        node.name = self.name.to_xml(factory)
        node.gender = self.gender.upper() if self.gender else "U"
        node.language = language_node
        node.email = self.email
        node.dateOfBirth = self.date_of_birth.isoformat() if self.date_of_birth else None   # yyyy-mm-dd
        node.phoneNumber = self.phone_number                # string50, must start with "+"
        node.mobilePhoneNumber = self.mobile_phone_number   # string50, must start with "+"
        node.ipAddress = self.ipAddress if self.ipAddress else None
        return node


class Destination(object):
    """
    Name and address to use for billing.
    """
    def __init__(self, name, address):
        """
        :type name: Name
        :type address: Address
        """
        self.name = name
        self.address = address

    @classmethod
    def from_address(cls, address):
        """
        :type address: :class:`~oscar.apps.address.abstract_models.AbstractAddress`
        """
        return cls(
            name=Name(
                first=address.first_name,
                last=address.last_name,
                prefix=address.title or None,
            ),
            address=Address.from_address(address),
        )

    def to_xml(self, factory):
        node = factory.create('ns0:destination')
        node.name = self.name.to_xml(factory)
        node.address = self.address.to_xml(factory)
        return node


class Address(object):
    """
    An address for docdata

    :type street: str
    :type house_number: str
    :type house_number_addition: str
    :type postal_code: str
    :type city: str
    :type state: str
    :type country_code: str
    :type company: str
    :type vatNumber: str
    :type careOf: str
    """
    def __init__(self, street, house_number, house_number_addition, postal_code, city, state, country_code, company=None, vatNumber=None, careOf=None):
        self.street = street
        self.house_number = house_number
        self.house_number_addition = house_number_addition
        self.postal_code = postal_code
        self.city = city
        self.state = state
        self.country_code = country_code

        self.company = company
        self.vatNumber = vatNumber
        self.careOf = careOf
        #self.kvkNummer    # rant: seriously? a Netherlands-specific field in the API?

    @classmethod
    def from_address(cls, address):
        """
        :type address: :class:`~oscar.apps.address.abstract_models.AbstractAddress`
        """
        # NOTE: oscar has no separate field for house number!
        # The field however is required! Could consider passing nbsp ('\xc2\xa0')
        house_number = "N/A"
        if appsettings.DOCDATA_HOUSE_NUMBER_FIELD:
            value = getattr(address, appsettings.DOCDATA_HOUSE_NUMBER_FIELD)
            if value and value.isdigit():
                house_number = text_type(int(value))

        return cls(
            street=address.line1[:32],       # Docdata has a 32 char limit on street
            house_number=house_number,
            house_number_addition=None,
            postal_code=address.postcode,
            city=address.city,
            state=address.state,
            country_code=address.country_id  # The Country.iso_3166_1_a2 field.
        )

    def to_xml(self, factory):
        country = factory.create('ns0:country')
        country._code = text_type(self.country_code)

        node = factory.create('ns0:address')
        node.street = text_type(self.street)
        node.houseNumber = text_type(self.house_number)  #string35
        node.houseNumberAddition = text_type(self.house_number_addition) if self.house_number_addition else None
        node.postalCode = text_type(self.postal_code.replace(' ', ''))  # Spaces aren't allowed in the Docdata postal code (type=NMTOKEN)
        node.city = text_type(self.city)
        node.state = text_type(self.state) if self.state else None
        node.country = country

        # Optional company info
        node.company = text_type(self.company) if self.company else None
        node.vatNumber = text_type(self.vatNumber) if self.vatNumber else None
        node.careOf = text_type(self.careOf) if self.careOf else None
        return node


class Amount(object):
    """
    An amount for Docdata.
    """
    def __init__(self, value, currency):
        self.value = value
        self.currency = currency

    def to_xml(self, factory):
        node = factory.create('ns0:amount')
        node.value = int(self.value * 100)   # No comma!
        node._currency = self.currency       # An attribute
        return node


class Vat(Amount):
    """
    An amount of Value-Added-Tax for an order.
    """
    def __init__(self, value, currency, rate):
        super(Vat, self).__init__(value, currency)
        self.rate = rate

    @classmethod
    def from_prices(cls, excl_tax, incl_tax, currency):
        """
        Calculate VAT by subtracting prices.
        """
        if excl_tax > incl_tax:
            raise ValueError("Vat.from_prices() requires excl_tax <= incl_tax")

        if excl_tax == incl_tax:
            # Though very impractical, be correct on the output instead of getting a division by zero error.
            rate = 0
        else:
            rate = int((D(incl_tax) / D(excl_tax) - 1) * 100)

        return cls(
            value=incl_tax - excl_tax,
            currency=currency,
            rate=rate,
        )

    def to_xml(self, factory):
        node = factory.create('ns0:vat')
        node.amount = super(Vat, self).to_xml(factory)
        node._rate = self.rate
        return node


class TotalVatAmount(Vat):
    """
    A variation of the Vat for the totalVatAmount array.
    """
    @classmethod
    def from_vat(cls, vat):
        """
        Create this element by copying the VAT object values.
        :type vat: Vat
        """
        return cls(
            value=vat.value,
            currency=vat.currency,
            rate=vat.rate,
        )

    def to_xml(self, factory):
        # Dear Docdata: apparently, reusing Vat was not possible..?
        #node = factory.create('ns0:totalVatAmount') does not support setting text.
        element = Element('ns0:totalVatAmount')
        element.setText(str(int(self.value * 100)))
        element.set('rate', self.rate)
        element.set('currency', self.currency)
        return element


class Quantity(object):
    """
    An quantity for Docdata
    """
    def __init__(self, value, unit='PCS'):
        """
        :param value: The numeric value
        :type value: int
        :param unit: The Unit of measurement.The attribute can have the following values:
                     PCS - pieces
                     SEC - seconds
                     BYT - bytes
                     KB - kilobytes
        :type unit: str
        """
        self.value = value
        self.unit = unit

    def to_xml(self, factory):
        # Needs to be an xsd:int with an attribute
        # Can't do that with factory.create('ns0:quantity')
        #metadata = factory.resolver.find('ns0:quantity')
        #ns = metadata.namespace()

        element = Element('ns0:quantity')
        element.setText(str(self.value))
        element.set('unitOfMeasure', self.unit)
        return element


class Invoice(object):
    """
    Optional additional data used for invoices.
    """
    def __init__(self, total_net_amount, total_vat_amount, items, ship_to, additional_description=None):
        """
        :type total_net_amount: Amount
        :type total_vat_amount: Vat
        :type items: list of Item
        :type ship_to: Destination
        :type additional_description: str
        """
        self.total_net_amount = total_net_amount
        self.total_vat_amount = total_vat_amount
        self.items = list(items) if items is not None else []
        self.ship_to = ship_to
        self.additional_description = additional_description

    @classmethod
    def from_basket(cls, basket, total, shipping_address):
        """
        :param basket: The basket to submit
        :type basket: :class:`~oscar.apps.basket.abstract_models.AbstractBasket`
        :param shipping_address: The shipping address for the invoice.
                                Note: the Docdata API v1.2 reads this address he "State" field to submit to PayPal.
                                Hence, HACK this by passing the billing address instead of the shipping address.
        :type shipping_address: :class:`~oscar.apps.address.abstract_models.AbstractAddress`
        """
        return cls(
            total_net_amount=Amount(basket.total_excl_tax, total.currency),
            total_vat_amount=Vat.from_prices(basket.total_excl_tax, basket.total_incl_tax, total.currency),
            additional_description="",
            items=[
                Item.from_line(line) for line in basket.lines.all()
                # TODO: Add shipping costs line!
            ],
            # Note: Docdata reads this field fort he "State" field to submit to PayPal.
            # Hence, hack this by passing the shipping_address instead of the billing address.
            ship_to=Destination.from_address(shipping_address)
        )

    def to_xml(self, factory):
        node = factory.create('ns0:invoice')
        node.totalNetAmount = self.total_net_amount.to_xml(factory)
        node.totalVatAmount = [TotalVatAmount.from_vat(self.total_vat_amount).to_xml(factory)]  # Can be repeated in WSDL...?!?
        node.item = [item.to_xml(factory) for item in self.items]
        node.shipTo = self.ship_to.to_xml(factory)
        if self.additional_description:
            node.additionalDescription = text_type(self.additional_description)  # max 100
        return node


class Item(object):
    """
    A single (line) item for this order.
    """
    def __init__(self, number, name, code, quantity, description, net_amount, gross_amount, vat, total_net_amount, total_gross_amount, total_vat, image_url=None):
        """
        :param number: Line ID
        :type number: long
        :param name: The human-readable name of this item. (max 50)
        :param code: A code or article number identifying this item. (max 50)
        :param quantity: Quantity of this item that's being ordered.
        :type quantity: Quantity
        :param description: The description of the item. (max 100)
        :param net_amount: The net amount for a single piece of this item.
        :type net_amount: Amount
        :param gross_amount: The gross amount for a single piece of this item.
        :type gross_amount: Amount
        :param vat: The gross amount for a single piece of this item.
        :type vat: Vat
        :param total_net_amount: The net amount for a single piece of this item.
        :type total_net_amount: Amount
        :param total_gross_amount: The gross amount for a single piece of this item.
        :type total_gross_amount: Amount
        :param total_vat: The gross amount for a single piece of this item.
        :type total_vat: Vat
        :param image_url: URL to the items image (max 2048)
        """
        # Support simple values too
        if isinstance(quantity, (int, long)):
            quantity = Quantity(quantity, unit='PCS')

        self.number = number
        self.name = name
        self.code = code
        self.quantity = quantity
        self.description = description
        self.net_amount = net_amount
        self.gross_amount = gross_amount
        self.vat = vat
        self.total_net_amount = total_net_amount
        self.total_gross_amount = total_gross_amount
        self.total_vat = total_vat
        self.image_url = image_url

    @classmethod
    def from_line(cls, line):
        """
        :type line: oscar.apps.basket.abstract_models.AbstractLine | oscar.apps.order.abstract_models.AbstractLine
        """
        from oscar.apps.basket.abstract_models import AbstractLine as BasketLine
        from oscar.apps.order.abstract_models import AbstractLine as OrderLine
        if isinstance(line, BasketLine):
            currency = line.price_currency
        elif isinstance(line, OrderLine):
            currency = line.order.currency
        else:
            raise TypeError("Invalid argument type: {0}".format(line.__class__.__name__))

        product = line.product
        return cls(
            number=line.id,
            name=product.get_title(),
            code=product.upc,
            quantity=Quantity(line.quantity),
            description=Truncator(product.description).chars(100),
            net_amount=Amount(line.unit_price_excl_tax, currency),
            gross_amount=Amount(line.unit_price_incl_tax, currency),
            vat=Vat.from_prices(line.unit_price_excl_tax, line.unit_price_incl_tax, currency),
            total_net_amount=Amount(line.line_price_excl_tax, currency),
            total_gross_amount=Amount(line.line_price_incl_tax, currency),
            total_vat=Vat.from_prices(line.line_price_excl_tax, line.line_price_incl_tax, currency),
            image_url=None,   # Can't use product.primary_image().original.url because Docdata doesn't generate thumbnails.
        )

    def to_xml(self, factory):
        node = factory.create('ns0:item')
        node._number = self.number
        node.name = text_type(self.name)
        node.code = text_type(self.code)
        node.quantity = self.quantity.to_xml(factory)
        node.description = text_type(self.description or '-')
        node.image = text_type(self.image_url) if self.image_url else None
        node.netAmount = self.net_amount.to_xml(factory)
        node.grossAmount = self.gross_amount.to_xml(factory)
        node.vat = self.vat.to_xml(factory)
        node.totalNetAmount = self.total_net_amount.to_xml(factory)
        node.totalGrossAmount = self.total_gross_amount.to_xml(factory)
        node.totalVat = self.total_vat.to_xml(factory)
        return node


class TechnicalIntegrationInfo(object):
    """
    Pass integration information to the API for debugging assistance.
    """
    def to_xml(self, factory):
        node = factory.create('ns0:technicalIntegrationInfo')
        node.webshopPlugin = "django-oscar-docdata"
        node.webshopPluginVersion = oscar_docdata_version
        node.programmingLanguage = "Python"
        return node


class Payment(object):
    """
    Base interface for all payment inputs.
    """
    payment_method = None
    request_parameter = None

    def to_xml(self, factory):
        raise NotImplementedError("Missing to_xml() implementation in {0}".format(self.__class__.__name__))


class AmexPayment(Payment):
    """
    American Express payment.
    """
    payment_method = DocdataClient.PAYMENT_METHOD_AMEX
    request_parameter = 'amexPaymentInput'

    def __init__(self, credit_card_number, expiry_date, cid, card_holder, email_address):
        self.credit_card_number = credit_card_number
        self.expiry_date = expiry_date
        self.cid = cid
        self.card_holder = card_holder
        self.email_address = email_address

    def to_xml(self, factory):
        node = factory.create('ns0:amexPaymentInput')
        node.creditCardNumber = self.credit_card_number
        node.expiryDate = self.expiry_date
        node.cid = self.cid
        node.cardHolder = text_type(self.card_holder)
        node.emailAddress = self.email_address
        return node


class MasterCardPayment(Payment):
    """
    Mastercard payment
    """
    payment_method = DocdataClient.PAYMENT_METHOD_MASTERCARD
    request_parameter = 'masterCardPaymentInput'

    def __init__(self, credit_card_number, expiry_date, cvc2, card_holder, email_address):
        self.credit_card_number = credit_card_number
        self.expiry_date = expiry_date
        self.cvc2 = cvc2
        self.card_holder = text_type(card_holder)
        self.email_address = email_address

    def to_xml(self, factory):
        node = factory.create('ns0:masterCardPaymentInput')
        node.creditCardNumber = self.credit_card_number
        node.expiryDate = self.expiry_date
        node.cvc2 = self.cvc2
        node.cardHolder = text_type(self.card_holder)
        node.emailAddress = self.email_address
        return node


class DirectDebitPayment(Payment):
    """
    Direct debit payment.
    """
    payment_method = DocdataClient.PAYMENT_METHOD_DIRECT_DEBIT
    request_parameter = 'directDebitPaymentInput'

    def __init__(self, holder_name, holder_city, holder_country_code, bic, iban):
        self.holder_name = holder_name
        self.holder_city = holder_city
        self.holder_country_code = holder_country_code
        self.bic = bic
        self.iban = iban

    def to_xml(self, factory):
        country = factory.create('ns0:country')
        country._code = self.holder_country_code

        node = factory.create('ns0:directDebitPaymentInput')
        node.holderName = text_type(self.holder_name)
        node.holderCity = text_type(self.holder_city)
        node.holderCountry = country
        node.bic = self.bic
        node.iban = self.iban
        return node


class IdealPayment(DirectDebitPayment):
    """
    Direct debit payment in The Netherlands.
    The visitor is redirected to the bank website where the payment is made,
    and then redirected back to the gateway.
    """
    payment_method = 'IDEAL'


class BankTransferPayment(Payment):
    """
    Bank transfer.
    https://support.docdatapayments.com/index.php?_m=knowledgebase&_a=viewarticle&kbarticleid=141&nav=0,7
    """
    payment_method = DocdataClient.PAYMENT_METHOD_BANK_TRANSFER
    request_parameter = 'bankTransferPaymentInput'

    def __init__(self, email_address):
        self.email_address = email_address

    def to_xml(self, factory):
        node = factory.create('ns0:bankTransferPaymentInput')
        node.emailAddress = self.email_address
        return node


class ElvPayment(Payment):
    """
    The German Electronic Direct Debit (Elektronisches Lastschriftverfahren or ELV)
    """
    payment_method = DocdataClient.PAYMENT_METHOD_ELV
    request_parameter = 'elvPaymentInput'

    def __init__(self, account_number, bank_code):
        self.account_number = account_number
        self.bank_code = bank_code

    def to_xml(self, factory):
        node = factory.create('ns0:elvPaymentInput')
        node.accountNumber = self.account_number
        node.bankCode = self.bank_code
        return node


def to_iso639_part1(language_code):
    """
    Convert codes like "en-us" to "en"
    """
    return language_code.split('-', 1)[0]



"""
The main code to interface with Docdata.

This module is Oscar agnostic, and can be used in any other project.
The Oscar specific code is in the facade.
"""
from datetime import timedelta
import logging
from decimal import Decimal as D
from django.core.exceptions import ImproperlyConfigured
from django.db import IntegrityError, transaction
from django.utils.timezone import now
from django.utils.translation import get_language
from oscar_docdata import appsettings
from oscar_docdata.exceptions import InvalidMerchant
from oscar_docdata.gateway import DocdataClient
from oscar_docdata.models import DocdataOrder, DocdataPayment
from oscar_docdata.signals import order_status_changed, payment_added, payment_updated


logger = logging.getLogger(__name__)


class Interface(object):
    """
    The methods to interface with the Docdata gateway.
    """

    # TODO: is this really needed?
    status_mapping = {
        DocdataClient.STATUS_NEW: DocdataOrder.STATUS_NEW,
        DocdataClient.STATUS_STARTED: DocdataOrder.STATUS_NEW,
        DocdataClient.STATUS_REDIRECTED_FOR_AUTHENTICATION: DocdataOrder.STATUS_IN_PROGRESS,
        DocdataClient.STATUS_AUTHORIZATION_REQUESTED: DocdataOrder.STATUS_PENDING,
        DocdataClient.STATUS_AUTHORIZED: DocdataOrder.STATUS_PENDING,
        DocdataClient.STATUS_PAID: DocdataOrder.STATUS_PENDING,  # Overwritten when it's totals are checked.

        DocdataClient.STATUS_CANCELLED: DocdataOrder.STATUS_CANCELLED,
        DocdataClient.STATUS_CHARGED_BACK: DocdataOrder.STATUS_CHARGED_BACK,
        DocdataClient.STATUS_CONFIRMED_PAID: DocdataOrder.STATUS_PAID,
        DocdataClient.STATUS_CONFIRMED_CHARGEDBACK: DocdataOrder.STATUS_CHARGED_BACK,
        DocdataClient.STATUS_CLOSED_SUCCESS: DocdataOrder.STATUS_PAID,
        DocdataClient.STATUS_CLOSED_CANCELLED: DocdataOrder.STATUS_CANCELLED,
    }


    def __init__(self, testing_mode=None, merchant_name=None, merchant_password=None):
        """
        Initialize the interface.
        If the testing_mode is not set, it defaults to the ``DOCDATA_TESTING`` setting.
        """
        if testing_mode is None:
            testing_mode = appsettings.DOCDATA_TESTING
        self.testing_mode = testing_mode
        self.client = DocdataClient(testing_mode, merchant_name=merchant_name, merchant_password=merchant_password)



    @classmethod
    def for_merchant(cls, merchant_name, testing_mode=None):
        """
        Generate the client with the proper credentials.
        This method is useful when there are multiple sub accounts in use.
        The proper account credentials are automatically selected
        from the ``DOCDATA_MERCHANT_PASSWORDS`` setting.
        :rtype: DocdataClient
        """
        try:
            password = appsettings.DOCDATA_MERCHANT_PASSWORDS[merchant_name]
        except KeyError:
            raise ImproperlyConfigured("No password provided in DOCDATA_MERCHANT_PASSWORDS for merchant '{0}'".format(merchant_name))

        return cls(
            testing_mode=testing_mode,
            merchant_name=merchant_name,
            merchant_password=password
        )


    def create_payment(self, order_number, total, user, language=None, description=None, profile=appsettings.DOCDATA_PROFILE, merchant_name=None, **kwargs):
        """
        Start a new payment session / container.

        This is the first step of any Docdata payment session.

        :param order_number: The order number generated by Oscar.
        :param total: The price object, including totals and currency.
        :type total: :class:`oscar.core.prices.Price`
        :type user: :class:`django.contrib.auth.models.User`
        :param language: The language to display the interface in.
        :param description
        :returns: The Docdata order reference ("order key").
        """
        if not language:
            language = get_language()

        if merchant_name is not None:
            client = DocdataClient.for_merchant(merchant_name=merchant_name, testing_mode=self.testing_mode)
        else:
            client = self.client

        # May raise an DocdataCreateError exception
        call_args = self.get_create_payment_args(
            # Pass all as kwargs, make it easier for subclasses to override using *args, **kwargs and fetch all by name.
            order_number=order_number,
            total=total,
            user=user,
            language=language,
            description=description,
            profile=profile,
            **kwargs
        )
        createsuccess = client.create(**call_args)

        # Track order_key for local logging
        destination = call_args.get('bill_to')
        self._store_create_success(
            merchant_name=str(client.merchant_name),
            order_number=order_number,
            order_key=createsuccess.order_key,
            amount=call_args['total_gross_amount'],
            language=language,
            country_code=destination.address.country_code if destination else None
        )

        # Return for further reference
        return createsuccess.order_key


    def get_create_payment_args(self, order_number, total, user, language=None, description=None, profile=appsettings.DOCDATA_PROFILE, **kwargs):
        """
        The arguments to pass to create a payment.
        This is implementation-specific, hence not implemented here.
        """
        raise NotImplementedError("Missing get_create_payment_args() implementation!")


    def _store_create_success(self, merchant_name, order_number, order_key, amount, language, country_code):
        """
        Store the order_key for local status checking.
        """
        DocdataOrder.objects.create(
            merchant_name=merchant_name,
            merchant_order_id=order_number,
            order_key=order_key,
            total_gross_amount=amount.value,
            currency=amount.currency,
            language=language,
            country=country_code
        )


    def get_payment_menu_url(self, request, order_key, return_url=None, client_language=None, **extra_url_args):
        """
        Return the URL to the payment menu,
        where the user can be redirected to after creating a successful payment.

        For more information, see :func:`DocdataClient.get_payment_menu_url`.
        """
        return self.client.get_payment_menu_url(request, order_key, return_url=return_url, client_language=client_language, **extra_url_args)


    def start_payment(self, order, payment, payment_method=None):
        """
        :type order: DocdataOrder
        """
        # Backwards compatibility fix, old parameter was named "order_key".
        if isinstance(order, basestring):
            order = DocdataOrder.objects.select_for_update().active_merchants().get(order_key=order)

        amount = None

        # This can raise an exception.
        client = DocdataClient.for_merchant(order.merchant_name, testing_mode=self.testing_mode)
        startsuccess = client.start(order.order_key, payment, payment_method=payment_method, amount=amount)

        self._set_status(order, DocdataOrder.STATUS_IN_PROGRESS)
        order.save()

        # Not updating the DocdataPayment objects here,
        # instead just wait for Docdata to call the status view.

        # Return for further reference.
        return startsuccess.payment_id


    def cancel_order(self, order):
        """
        Cancel the order.
        :type order: DocdataOrder
        """
        client = DocdataClient.for_merchant(order.merchant_name, testing_mode=self.testing_mode)
        client.cancel(order.order_key)  # Can bail out with an exception (already logged)

        # Don't wait for server to send event back, get most recent state now.
        # Also make sure the order will be marked as cancelled.
        statusreply = client.status(order.order_key)  # Can bail out with an exception (already logged)
        self._store_report(order, statusreply.report, indented_status=DocdataOrder.STATUS_CANCELLED)


    def update_order(self, order):
        """
        :type order: DocdataOrder
        """
        # Fetch the latest status
        client = DocdataClient.for_merchant(order.merchant_name, testing_mode=self.testing_mode)
        if client.merchant_name != order.merchant_name:
            raise InvalidMerchant("Order {0} belongs to a different merchant: {1} (client uses: {2})".format(
                order.merchant_order_id, order.merchant_name, client.merchant_name
            ))

        statusreply = client.status(order.order_key)  # Can bail out with an exception (already logged)

        # Store the new status
        self._store_report(order, statusreply.report)


    def _store_report(self, order, report, indented_status=None):
        """
        Store the retrieved status report in the order object.

        :type order: DocdataOrder
        """
        # Store totals
        totals = report.approximateTotals
        order.total_registered = D(totals.totalRegistered) / 100
        order.total_shopper_pending = D(totals.totalShopperPending) / 100
        order.total_acquirer_pending = D(totals.totalAcquirerPending) / 100
        order.total_acquirer_approved = D(totals.totalAcquirerApproved) / 100
        order.total_captured = D(totals.totalCaptured) / 100
        order.total_refunded = D(totals.totalRefunded) / 100
        order.total_charged_back = D(totals.totalChargedback) / 100

        if hasattr(report, 'payment'):
            # Store all report lines, make an analytics of the new status
            new_status, ddpayments = self._store_report_lines(order, report)
        else:
            # There are no payments. It's really annoying to see that the Docdata status API
            # doesn't actually return a global "payment cluster" status code.
            # There are only status codes for the payment (which corresponds with a payment attempts by the user).
            # Make our best efforts here, based on some heuristics of the approximateTotals field.
            if totals.totalShopperPending == 0 \
            and totals.totalAcquirerPending == 0 \
            and totals.totalAcquirerApproved == 0 \
            and totals.totalCaptured == 0 \
            and totals.totalRefunded == 0 \
            and totals.totalChargedback == 0:
                # Everything is 0, either started, cancelled or expired
                if order.status == DocdataOrder.STATUS_CANCELLED:
                    new_status = order.status  # Stay in cancelled, don't become expired
                else:
                    if order.created < (now() - timedelta(days=21)):
                        # Will only expire old orders of more then 21 days.
                        new_status = indented_status or DocdataOrder.STATUS_EXPIRED
                    else:
                        # Either new or cancelled, can't determine!
                        new_status = indented_status or DocdataOrder.STATUS_NEW
            else:
                logger.error(
                    "Payment cluster %s has no payment yet, and unknown 'approximateTotals' heuristics.\n"
                    "Status can't be reliably determined. Please investigate.\n"
                    "Totals=%s", order.order_key, totals
                )
                if order.status in (DocdataOrder.STATUS_EXPIRED, DocdataOrder.STATUS_CANCELLED):
                    # Stay in cancelled/expired, don't switch back to NEW
                    new_status = order.status
                else:
                    new_status = indented_status or DocdataOrder.STATUS_NEW

        # Store status
        old_status = order.status
        status_changed = self._set_status(order, new_status)
        order.save()

        if status_changed:
            self.order_status_changed(order, old_status, order.status)


    def _set_status(self, order, new_status):
        """
        Changes the payment status to new_status and sends a signal about the change.
        """
        old_status = order.status
        if old_status != new_status:
            if new_status not in dict(DocdataOrder.STATUS_CHOICES):
                new_status = DocdataOrder.STATUS_UNKNOWN
                logger.warning("Payment cluster {0} status changed {1} -> {2} -> UNKNOWN!".format(order.order_key, old_status, new_status))
            else:
                logger.info("Payment cluster {0} status changed {1} -> {2}".format(order.order_key, old_status, new_status))

            order.status = new_status
            return True
        else:
            return False


    def _store_report_lines(self, order, report):
        """
        Store the status report lines from the StatusReply.
        Each line represents a payment event, which is stored in a DocdataPayment object.

        This performs the checks related to the status change.
        This returns the "status" value of the last payment line.
        This line either indicates the payment is authorized, cancelled, refunded, etc..

        :type order: DocdataOrder
        """
        new_status = None
        ddpayment_objects = []
        totals = report.approximateTotals

        logger.info("Payment cluster {0} Total Registered: {1} Total Captured: {2} Total Chargedback: {3} Total Refunded: {4}".format(
            order.order_key, totals.totalRegistered, totals.totalCaptured, totals.totalChargedback, totals.totalRefunded
        ))

        # Webservice doesn't return payments in the correct order (or reversed).
        # So far, the payments can only be sorted by ID.
        report_payments = list(report.payment)
        report_payments.sort(key=lambda payments: payments.id)

        for payment in report_payments:
            # payment_report is a ns0:payment object, which contains:
            # - id            (paymentId, a positiveInteger)
            # - paymentMethod (string50)
            # - authorization  (authorization)
            #   - status      str
            #   - amount      (amount); value + currency attribute.
            #   - confidenceLevel  (string35)
            #   - capture     (capture); status, amount, reason
            #   - refund      (refund); status, amount, reason
            #   - chargeback  (chargeback); status, amount, reason
            # - extended      payment specific information, depends on payment method.

            logger.debug("- Payment {0} with {1}: auth status: {2}".format(payment.id, payment.paymentMethod, payment.authorization.status))

            authorization = payment.authorization
            auth_status = str(payment.authorization.status)

            if auth_status == 'AUTHORIZED':
                # The payment was authorized, check what the contents of it is.
                # This validates the status, and determines which amount got paid.
                maybe_new_status = self._process_authorized_payment(order, report, payment)
                if maybe_new_status is not None:
                    new_status = maybe_new_status

                # NOTE: currencies ignored here.
                # This only indicates the amount that's being dealt with.
                # the actual debited value is added when the value is captured.
                amount_allocated = _to_decimal(authorization.amount)
            else:
                amount_allocated = 0


            # Now save the result into a DocdataPayment object.
            # Find or create the correct payment object for current report.
            payment_class = DocdataPayment #TODO: self.id_to_model_mapping[order.payment_method_id]
            updated = False
            added = False

            try:
                ddpayment = payment_class.objects.select_for_update().get(payment_id=str(payment.id))
            except payment_class.DoesNotExist:
                # Create new line
                ddpayment = payment_class(
                    docdata_order=order,
                    payment_id=long(payment.id),
                    payment_method=str(payment.paymentMethod),
                )
                added = True

            if not payment.paymentMethod == ddpayment.payment_method:
                # Payment method change??
                logger.warn(
                    "Payment method from Docdata doesn't match saved payment method. "
                    "Storing the payment method received from Docdata for payment id {0}: {1}".format(
                        ddpayment.payment_id, payment.paymentMethod
                ))
                ddpayment.payment_method = str(payment.paymentMethod)
                updated = True

            # Store the totals
            old_values = (ddpayment.confidence_level, ddpayment.amount_allocated, ddpayment.amount_chargeback, ddpayment.amount_refunded, ddpayment.amount_debited)

            ddpayment.confidence_level = authorization.confidenceLevel
            ddpayment.amount_allocated = amount_allocated
            ddpayment.amount_debited = self._get_payment_sum(payment, "capture", "CAPTURED")
            ddpayment.amount_refunded = self._get_payment_sum(payment, "refund", "CAPTURED")
            ddpayment.amount_chargeback = self._get_payment_sum(payment, "chargeback", "CHARGED")

            # Track changes
            new_values = (ddpayment.confidence_level, ddpayment.amount_allocated, ddpayment.amount_chargeback, ddpayment.amount_refunded, ddpayment.amount_debited)
            if old_values != new_values:
                updated = True

            # Detect status change

            if ddpayment.status != auth_status:
                # Status change!
                logger.info("Docdata payment status changed. payment={0} status: {1} -> {2}".format(
                    payment.id, ddpayment.status, auth_status
                ))

                if auth_status not in DocdataClient.DOCUMENTED_STATUS_VALUES \
                and auth_status not in DocdataClient.SEEN_UNDOCUMENTED_STATUS_VALUES:
                    # Note: We continue to process the payment status change on this error.
                    logger.warn("Received unknown payment status from Docdata. payment={0}, status={1}".format(
                        payment.id, auth_status
                    ))

                ddpayment.status = auth_status
                updated = True

            if added or updated:
                # Saving might happen concurrently, as the user returns to the OrderReturnView
                # and Docdata calls the StatusChangedNotificationView at the same time.
                sid = transaction.savepoint()  # for PostgreSQL
                try:
                    ddpayment.save()
                    transaction.savepoint_commit(sid)
                except IntegrityError:
                    transaction.savepoint_rollback(sid)
                    logger.warn("Experienced concurrency issues with update-status, payment id {0}: {1}".format(payment.id))

                    # Overwrite existing object instead.
                    #not needed, no impact on save: ddpayment._state.adding = False
                    ddpayment.id = str(payment.id)
                    ddpayment.save()
                    added = False

                # Fire events so payment transactions can be created in Oscar.
                # This can be used to call source.transactions.create(..) for example.
                if added:
                    payment_added.send(sender=DocdataPayment, order=order, payment=ddpayment)
                else:
                    payment_updated.send(sender=DocdataPayment, order=order, payment=ddpayment)

            ddpayment_objects.append(ddpayment)
            setattr(ddpayment, '_source', payment)


        if new_status is None:
            # Didn't get a clearly detectable/conclusive status.
            # Try to use the last line in such case, otherwise, use new_status.
            #
            # This handles the strange situation we've seen:
            # - Customer initiated both a PayPal and VISA payment
            # - Then completes the PayPal payment.
            # - Hence the last payment is NEW, but the first is AUTHORIZED.

            # Some status mapping overrides.
            new_status = self.status_mapping.get(report_payments[-1].authorization.status, DocdataOrder.STATUS_UNKNOWN)

            # Stay in cancelled/expired, don't switch back to NEW
            # Even though the payment cluster is set to 'closed_expired',
            # Docdata doesn't expire the individual payment report lines.
            if order.status in (DocdataOrder.STATUS_EXPIRED, DocdataOrder.STATUS_CANCELLED) \
            and new_status in (DocdataOrder.STATUS_NEW, DocdataOrder.STATUS_IN_PROGRESS):
                new_status = order.status

            # TODO Use status change log to investigate if these overrides are needed.
            # # These overrides are really just guessing.
            # latest_capture = authorization.capture[-1]
            # if status == 'AUTHORIZED':
            #     if hasattr(authorization, 'refund') or hasattr(authorization, 'chargeback'):
            #         new_status = 'cancelled'
            #     if latest_capture.status == 'FAILED' or latest_capture == 'ERROR':
            #         new_status = 'failed'
            #     elif latest_capture.status == 'CANCELLED':
            #         new_status = 'cancelled'

        # Detect a nasty error condition that needs to be manually fixed.
        total_registered = long(totals.totalRegistered)
        total_gross_cents = long(order.total_gross_amount * 100)
        if new_status != DocdataOrder.STATUS_CANCELLED and total_registered != total_gross_cents:
            logger.error("Payment cluster %s total: %s does not equal Total Registered: %s.",
                order.order_key, total_gross_cents, total_registered
            )

        # Webservice doesn't return payments in the correct order (or reversed).
        # So far, the payments can only be sorted by ID.
        ddpayment_objects.sort(key=lambda ddpayment: ddpayment.payment_id)
        return new_status, ddpayment_objects


    def _get_payment_sum(self, payment, xml_tag, success_status):
        """
        Take the sum of multiple <capture>, <refund> or <chargeback> elements.
        """
        amount = D("0.00")
        authorization = payment.authorization
        if hasattr(authorization, xml_tag):
            # There was some income/refund/chargeback
            for tag in getattr(authorization, xml_tag):
                if tag.status == success_status:
                    amount += _to_decimal(tag.amount)
                else:
                    logger.debug("{0} of {1} is marked as {2}, not adding to totals".format(tag.__class__.__name__.title(), payment.id, tag.status))

        return amount


    def _process_authorized_payment(self, order, report, payment):
        """
        Process the "authorization" block in a single payment.
        This tells whether the payment object was a capture, refund or chargeback.
        The expected totals are compared for accuracy.

        The new_status could remain None.
        A value is only returned when there is a clearly detectable status.

        :type order: DocdataOrder
        :rtype: str|None
        """
        totals = report.approximateTotals
        new_status = None

        # Because currency conversions may cause payments to happen with a few cents less,
        # this workaround makes sure those orders will still be marked as paid!
        # If you don't like this, the alternative is using DOCDATA_PAYMENT_SUCCESS_MARGIN = {}
        # and listening for the callback=SUCCESS value in the `return_view_called` signal.
        margin = 0
        if order.currency == totals._exchangedTo:  # Reads XML attribute.
            if any(p.authorization.amount._currency != order.currency for p in report.payment):
                # Order has a currency conversion, apply the margin
                margin = appsettings.DOCDATA_PAYMENT_SUCCESS_MARGIN.get(totals._exchangedTo, 0)

                # But if it exceeds the totalRegistered (e.g. it's 0), avoid making everything as paid!
                if margin >= totals.totalRegistered:
                    margin = 0


        # Integration Manual Order API 1.0 - Document version 1.0, 08-12-2012 - Page 33:
        #
        # Safe route: The safest route to check whether all payments were made is for the merchants
        # to refer to the "Total captured" amount to see whether this equals the "Total registered
        # amount". While this may be the safest indicator, the downside is that it can sometimes take a
        # long time for acquirers or shoppers to actually have the money transferred and it can be
        # captured.
        #
        if totals.totalCaptured < (totals.totalRegistered - margin):
            return None


        # The single payment indicated there is a payment.
        # Now comparing the totals, to see whether the order was fully paid!
        payment_sum = (totals.totalCaptured - totals.totalChargedback - totals.totalRefunded)

        if payment_sum >= (totals.totalRegistered - margin):
            # With all capture changes etc.. it's still what was registered.
            # Full amount is paid.
            new_status = DocdataOrder.STATUS_PAID
            logger.info("Payment cluster {0} Total Registered: {1} >= Captured: {2} (margin: {3}); new status PAID".format(
                order.order_key, totals.totalRegistered, totals.totalCaptured, margin
            ))

        elif payment_sum == 0:
            # A payment was captured, but the totals are 0.
            # See if there is a charge back or refund.

            # See what happened with the last payment addition
            authorization = payment.authorization

            # Example data:
            #
            # <payment>
            #     <id>2530366542</id>
            #     <paymentMethod>AMEX</paymentMethod>
            #     <authorization>
            #         <status>AUTHORIZED</status>
            #         <amount currency="USD">23700</amount>
            #         <confidenceLevel>ACQUIRER_APPROVED</confidenceLevel>
            #         <capture>
            #             <status>CAPTURED</status>
            #             <amount currency="USD">23700</amount>
            #         </capture>
            #         <chargeback>
            #             <chargebackId>437055</chargebackId>
            #             <status>CHARGED</status>
            #             <amount currency="USD">23700</amount>
            #         </chargeback>
            #     </authorization>
            # </payment>
            #
            # There can be multiple capture and chargeback objects.

            # Chargeback.
            # TODO: Add chargeback fee somehow (currently E0.50).
            if totals.totalCaptured == totals.totalChargedback:
                if hasattr(authorization, 'chargeback') and len(authorization.chargeback) > 0:
                    for chargeback in authorization.chargeback:
                        reason = getattr(chargeback, 'reason', '(reason not provided)')
                        logger.info("- Payment {0} chargedback: {1} {2}, {3}".format(
                            payment.id, chargeback.amount._currency, chargeback.amount.value, reason
                        ))
                else:
                    logger.info("Payment cluster {0} chargedback.".format(order.order_key))

                new_status = DocdataOrder.STATUS_CHARGED_BACK

            # Refund.
            # TODO: Log more info from refund when we have an example.
            if totals.totalCaptured == totals.totalRefunded:
                logger.info("Payment cluster {0} refunded.".format(order.order_key))
                new_status = DocdataOrder.STATUS_REFUNDED
        elif payment_sum > 0:
            # There is a partial refund.
            new_status = DocdataOrder.STATUS_PAID_REFUNDED

            logger.info("Payment cluster {0} Total Registered: {1} < Captured: {2} - Refunded: {3} - Chargeback: {4}  (margin: {5}); new status PAID_REFUNDED".format(
                order.order_key, totals.totalRegistered, totals.totalCaptured, totals.totalRefunded, totals.totalChargedback, margin
            ))

        else:
            # Show as error instead, this is not handled yet.
            logger.error(
                "Payment cluster %s chargeback and refunded sum is negative. Please investigate.\n"
                "Payment sum=%s Totals=%s", order.order_key, payment_sum, totals
            )
            new_status = DocdataOrder.STATUS_UNKNOWN

        return new_status


    def order_status_changed(self, docdataorder, old_status, new_status):
        """
        Notify that the order status changed.
        This function can be extended by inheriting the Facade class.
        """
        if old_status == new_status:
            return

        # Note that using a custom Facade class in your project doesn't help much,
        # as the Facade is also used by the default views.
        order_status_changed.send(sender=DocdataOrder, order=docdataorder, old_status=old_status, new_status=new_status)


def _to_decimal(amount):
    # Convert XML amount to decimal
    return D(long(amount.value)) / 100



from django.conf.urls import url
from .views import OrderReturnView, StatusChangedNotificationView


urlpatterns = [
    url(r'^return/$', OrderReturnView.as_view(), name='return_url'),
    url(r'^update_order/$', StatusChangedNotificationView.as_view(), name='status_changed'),
]



import logging
from django.http import HttpResponseBadRequest, HttpResponseRedirect, HttpResponse, HttpResponseNotFound, Http404, \
    HttpResponseServerError
from django.utils import translation
from django.views.generic import View
from oscar_docdata import appsettings
from oscar_docdata.compat import transaction_atomic
from oscar_docdata.exceptions import DocdataStatusError
from oscar_docdata.facade import get_facade
from oscar_docdata.models import DocdataOrder
from oscar_docdata.signals import return_view_called, status_changed_view_called

logger = logging.getLogger(__name__)


class UpdateOrderMixin(object):
    """
    Base mixin for updating the order from the API or return URL.
    """

    # What docdata calls the order_id, we call the order_key
    # Docdata uses both the order_key and merchant_order_id in the requests, depending on the view.
    order_query_arg = 'order_id'
    order_slug_field = 'order_key'

    # This is still here for backwards compatibility,
    # but setting ``DOCDATA_FACADE_CLASS`` is a better option nowadays.
    facade_class = None

    def get_facade(self):
        if self.facade_class is not None:
            return self.facade_class()
        else:
            # Making sure the object is lazy-imported.
            return get_facade()

    def get_order_slug(self):
        try:
            return self.request.GET[self.order_query_arg]
        except KeyError:
            raise KeyError("Missing {0} parameter".format(self.order_query_arg))

    def get_order(self, order_slug):
        """
        Update the status of an order, by fetching the latest state from docdata.
        """
        # Try to find the order.
        # Block any other requests on the order until the status change is handled.
        try:
            return DocdataOrder.objects.select_for_update().active_merchants().get(**{self.order_slug_field: order_slug})
        except DocdataOrder.DoesNotExist:
            logger.error("Order {0}='%s' not found to update payment status.".format(self.order_slug_field), order_slug)
            raise Http404(u"Order {0}='{1}' not found!".format(self.order_slug_field, order_slug))

    def update_order(self, order):
        # Ask the facade to request the status, and update the order accordingly.
        facade = self.get_facade()
        facade.update_order(order)


class OrderReturnView(UpdateOrderMixin, View):
    """
    The view to redirect to after a successful order creation.

    The URL to this view can be entered in the Docdata backoffice.
    Include the full URL, including ``?callback=...&order_id=``. The URL is appended with the order ID.
    For example::

        http://www.merchantwebsite.com/api/docdata/return/?callback=SUCCESS&order_id=
        http://www.merchantwebsite.com/api/docdata/return/?callback=CANCELLED&order_id=
        http://www.merchantwebsite.com/api/docdata/return/?callback=ERROR&order_id=
        http://www.merchantwebsite.com/api/docdata/return/?callback=PENDING&order_id=

    When using :func:`~oscar_docdata.gateway.Gateway.get_payment_menu_url`, these
    URLs are also included in the redirect URL. However, the Docdata Backoffice
    requires default URLs to be filled in.

    The use of this service is optional, but recommended.
    """
    order_slug_field = 'order_key'
    success_url = appsettings.DOCDATA_SUCCESS_URL
    pending_url = appsettings.DOCDATA_PENDING_URL
    cancelled_url = appsettings.DOCDATA_CANCELLED_URL
    error_url = appsettings.DOCDATA_ERROR_URL

    def get(self, request, *args, **kwargs):
        # Directly query the latest state from Docdata
        try:
            order_key = self.get_order_slug()
        except KeyError as e:
            return HttpResponseBadRequest(e.message, content_type='text/plain; charset=utf-8')

        callback = request.GET.get('callback') or ''
        logger.info("Returned from Docdata for {0}, callback: {1}".format(order_key, callback))

        # Need to make sure the latest status is present,
        # won't wait for Docdata to call our update API.
        with transaction_atomic():
            self.order = self.get_order(order_key)   # this is the docdata id.
            self.update_order(self.order)

        # Allow other code to perform actions, e.g. send a confirmation email.
        responses = return_view_called.send(sender=self.__class__, request=request, order=self.order, callback=callback)

        # Allow a signal receiver to override the returned URL and return a response.
        for receiver, response in responses:
            if response and isinstance(response, HttpResponse):
                return response

        # Redirect to thank you page, or the cancelled page,
        # depending on the callback parameter
        with translation.override(self.order.language):   # Allow i18n_patterns() to work properly
            url = str(self.get_redirect_url(callback))    # force evaluation of reverse_lazy()

        return HttpResponseRedirect(url)

    def get_redirect_url(self, callback):
        """
        Return the URL to redirect to.
        The callback value can be 'SUCCESS', 'PENDING', 'CANCELLED', 'ERROR'
        """
        # The callback parameter is added by get_payment_menu_url()
        if callback == 'SUCCESS':
            return self.success_url
        elif callback in ('PENDING', ''):  # treat missing value as pending
            return self.pending_url
        elif callback == 'CANCELLED':
            return self.cancelled_url
        else:
            return self.error_url


class StatusChangedNotificationView(UpdateOrderMixin, View):
    """
    Status changed notification view.

    The URL to this view can be entered in the Docdata backoffice.
    Include the full URL, including ``?order_id=``. The URL is appended with the order ID.
    For example:

    ``http://www.merchantwebsite.com/api/docdata/update_order/?order_id=``

    The use of this service is optional, but recommended.
    """
    order_slug_field = 'merchant_order_id'

    def get(self, request, *args, **kwargs):
        try:
            order_key = self.get_order_slug()
        except KeyError as e:
            return HttpResponseBadRequest(e.message, content_type='text/plain; charset=utf-8')

        logger.info("Got Docdata status changed notification for {0}".format(order_key))

        with transaction_atomic():
            try:
                self.order = self.get_order(order_key)  # Inconsistent, this call uses the merchant_order_id
            except Http404 as e:
                return HttpResponseNotFound(str(e), content_type='text/plain; charset=utf-8')

            try:
                self.update_order(self.order)
            except DocdataStatusError as e:
                logger.exception("The order status could not be retrieved from Docdata by the notification-url")
                return HttpResponseServerError(
                    "Failed to fetch status from Docdata API.\n"
                    "\n\n"
                    "Docdata API response:\n"
                    "---------------------\n"
                    "\n"
                    "code:    {0}\n"
                    "message: {1}".format(e.code, e.message),
                    content_type='text/plain; charset=utf-8'
                )

        responses = status_changed_view_called.send(sender=self.__class__, request=request, order=self.order)

        # Return 200 as required by DocData when the status changed notification was consumed.
        return HttpResponse(u"ok, order {0} updated\n".format(order_key), content_type='text/plain; charset=utf-8')



import traceback
import sys
from django.core.exceptions import ImproperlyConfigured


def import_class(import_path, setting_name=None):
    """
    Import a class by name.
    """
    mod_name, class_name = import_path.rsplit('.', 1)

    # import module
    mod = _import_module(mod_name, classnames=(class_name,))
    if mod is not None:
        # Loaded module, get attribute
        try:
            return getattr(mod, class_name)
        except AttributeError:
            pass

    # For ImportError and AttributeError, raise the same exception.
    if setting_name:
        raise ImproperlyConfigured("{0} does not point to an existing class: {1}".format(setting_name, import_path))
    else:
        raise ImproperlyConfigured("Class not found: {0}".format(import_path))


# Based on code from django-oscar:
def _import_module(module_label, classnames=()):
    """
    Imports the module with the given name.
    Returns None if the module doesn't exist, but propagates any import errors.
    """
    try:
        return __import__(module_label, fromlist=classnames)
    except ImportError:
        # There are 2 reasons why there could be an ImportError:
        #
        #  1. Module does not exist. In that case, we ignore the import and return None
        #  2. Module exists but another ImportError occurred when trying to import the module.
        #     In that case, it is important to propagate the error.
        #
        # ImportError does not provide easy way to distinguish those two cases.
        # Fortunately, the traceback of the ImportError starts at __import__
        # statement. If the traceback has more than one frame, it means that
        # application was found and ImportError originates within the local app
        __, __, exc_traceback = sys.exc_info()
        frames = traceback.extract_tb(exc_traceback)
        if len(frames) > 1:
            raise
    return None



"""
Django compatibility features
"""

__all__ = (
    'transaction_atomic',
    'get_model',
)

# New transaction support in Django 1.6
from django.db import transaction
try:
    transaction_atomic = transaction.atomic
except AttributeError:
    transaction_atomic = transaction.commit_on_success

# Oscar 0.7 get_model code
try:
    from oscar.core.loading import get_model
except ImportError:
    from django.db.models import get_model



from django import forms
from django.utils.translation import ugettext_lazy as _
from oscar_docdata.models import DocdataOrder


class DocdataOrderSearchForm(forms.Form):
    order_number = forms.CharField(required=False, label=_("Order number"))
    order_status = forms.ChoiceField(required=False, label=_("Status"), choices=(('', '---------'),) + DocdataOrder.STATUS_CHOICES)



from . import appsettings
from django.db import models
from django.db.models.query import QuerySet
from six import iterkeys
import django


class DocdataOrderQuerySet(QuerySet):
    """
    Chainable methods
    """
    def active_merchants(self):
        """
        Only select the docdata accounts of ``DOCDATA_MERCHANT_PASSWORDS``
        """
        if len(appsettings.DOCDATA_MERCHANT_PASSWORDS) == 1:
            # Optimize SQL a little bit
            return self.filter(merchant_name=next(iterkeys(appsettings.DOCDATA_MERCHANT_PASSWORDS)))
        else:
            return self.filter(merchant_name__in=appsettings.DOCDATA_MERCHANT_PASSWORDS.keys())

    def current_merchant(self):
        """
        Only select the current docdata account.
        """
        return self.filter(merchant_name=appsettings.DOCDATA_MERCHANT_NAME)

    def for_reference(self, docdata_ref):
        """
        Find an order by the reference that create_payment() returns (the "order key")
        """
        return self.get(order_key=docdata_ref)

    def for_order(self, order_number):
        """
        Find an order by the Oscar order number.
        """
        return self.get(merchant_order_id=order_number)


class DocdataOrderManager(models.Manager):
    """
    Extra methods for the DocdataOrder model.
    """
    queryset_class = DocdataOrderQuerySet

    def get_queryset(self):
        return self.queryset_class(self.model, using=self._db)

    if django.VERSION < (1,6):
        # For Django 1.5
        def get_query_set(self):
            return self.queryset_class(self.model, using=self._db)

    def active_merchants(self):
        """
        Only select the orders
        """
        # using .all() so Django selects the proper get_queryset() method.
        return self.all().active_merchants()

    def current_merchant(self):
        """
        Only select the current docdata account.
        """
        return self.all().current_merchant()

    def for_reference(self, docdata_ref):
        """
        Find an order by the reference that create_payment() returns (the "order key")
        """
        return self.all().for_reference(docdata_ref)

    def for_order(self, order_number):
        """
        Find an order by the Oscar order number.
        """
        return self.all().for_order(order_number)



"""
Settings for this application
"""
from django.conf import settings

# Credentials as supplied by the payment provider.
from django.core.urlresolvers import reverse_lazy

DOCDATA_MERCHANT_NAME = getattr(settings, 'DOCDATA_MERCHANT_NAME')
DOCDATA_MERCHANT_PASSWORD = getattr(settings, 'DOCDATA_MERCHANT_PASSWORD')
DOCDATA_MERCHANT_PASSWORDS = getattr(settings, 'DOCDATA_MERCHANT_PASSWORDS', {})

# Add the common case
if DOCDATA_MERCHANT_NAME not in DOCDATA_MERCHANT_PASSWORDS:
    DOCDATA_MERCHANT_PASSWORDS[DOCDATA_MERCHANT_NAME] = DOCDATA_MERCHANT_PASSWORD

# Whether to use the testing mode, or live mode.
DOCDATA_TESTING = getattr(settings, 'DOCDATA_TESTING', True)

# The default profile that is used to select the available payment methods that can be used to pay this order.
# Note this profile needs to be created in the Docdata Backoffice. By default, there is no profile available,
# but the parameter is required in the API call.
DOCDATA_PROFILE = getattr(settings, 'DOCDATA_PROFILE', 'standard')

# The default expected number of days in which the payment should be processed, or be expired if not paid.
# This appears to be totally ignored by Docdata, choosing their default 21 days (based on manual testing)
DOCDATA_DAYS_TO_PAY = getattr(settings, 'DOCDATA_DAYS_TO_PAY', 7)

# The default URL to redirect to. Defaults to a django-oscar view, but it can be any view of your choice.
DOCDATA_REDIRECT_URL = getattr(settings, 'DOCDATA_REDIRECT_URL', reverse_lazy('checkout:thank-you'))

DOCDATA_SUCCESS_URL = getattr(settings, 'DOCDATA_SUCCESS_URL', DOCDATA_REDIRECT_URL)
DOCDATA_PENDING_URL = getattr(settings, 'DOCDATA_PENDING_URL', DOCDATA_REDIRECT_URL)
DOCDATA_ERROR_URL = getattr(settings, 'DOCDATA_ERROR_URL', '/')
DOCDATA_CANCELLED_URL = getattr(settings, 'DOCDATA_CANCELLED_URL', '/')

# Facade class to use
DOCDATA_FACADE_CLASS = getattr(settings, 'DOCDATA_FACADE_CLASS', 'oscar_docdata.facade.Facade')

# Field to read for the housenumber field. (e.g. line3?)
DOCDATA_HOUSE_NUMBER_FIELD = getattr(settings, 'DOCDATA_HOUSE_NUMBER_FIELD', None)

# Translate the docdata order status to the configured OSCAR_ORDER_STATUS_PIPELINE
# If a value is missing, the raw DocdataOrder status value will be inserted.
# Possible values are: new, in_progress, paid, changed_back, cancelled, pending, refunded, unknown
DOCDATA_ORDER_STATUS_MAPPING = getattr(settings, 'DOCDATA_ORDER_STATUS_MAPPING', {})

# Native oscar setting:
OSCAR_ORDER_STATUS_CASCADE = getattr(settings, 'OSCAR_ORDER_STATUS_CASCADE', {})

# Allow 100 cents difference between the total received payment
# and actual registered money to handle currency rate conversions.
DOCDATA_PAYMENT_SUCCESS_MARGIN = getattr(settings, 'DOCDATA_PAYMENT_SUCCESS_MARGIN', {
    'USD': 160,
    'EUR': 100,
})

# The list of known issuers (banks) for iDEAL
DOCDATA_IDEAL_ISSUERS = getattr(settings, 'DOCDATA_IDEAL_ISSUERS', {
    '0081': 'Fortis',
    '0021': 'Rabobank',
    '0721': 'ING Bank',
    '0751': 'SNS Bank',
    '0031': 'ABN Amro Bank',
    '0761': 'ASN Bank',
    '0771': 'SNS Regio Bank',
    '0511': 'Triodos Bank',
    '0091': 'Friesland Bank',
    '0161': 'van Lanschot Bankiers'
})



from __future__ import unicode_literals
from django.utils.encoding import python_2_unicode_compatible


from decimal import Decimal as D
from django.db import models
from django.utils.translation import ugettext_lazy as _
from oscar_docdata.managers import DocdataOrderManager
from . import appsettings

try:
    from polymorphic.models import PolymorphicModel  # django-polymorphic 0.8
except ImportError:
    from polymorphic import PolymorphicModel


@python_2_unicode_compatible
class DocdataOrder(models.Model):
    """
    Tracking of the order which is sent to docdata.
    """
    # Simplified internal status codes.
    # Lowercased on purpose to avoid mixing the statuses together.
    STATUS_NEW = 'new'                     # Initial state
    STATUS_IN_PROGRESS = 'in_progress'     # In the redirect phase
    STATUS_PENDING = 'pending'             # Waiting for user to complete payment (e.g. credit cards)
    STATUS_PAID = 'paid'                   # End of story, paid!
    STATUS_PAID_REFUNDED = 'paid_refunded' # Paid, and performed a partial refund
    STATUS_CANCELLED = 'cancelled'         # End of story, cancelled
    STATUS_CHARGED_BACK = 'charged_back'   # End of story, consumer asked for charge back
    STATUS_REFUNDED = 'refunded'           # End of story, refunded, merchant refunded
    STATUS_EXPIRED = 'expired'             # No results of customer, order was closed.
    STATUS_UNKNOWN = 'unknown'             # Help!

    STATUS_CHOICES = (
        (STATUS_NEW, _("New")),
        (STATUS_IN_PROGRESS, _("In Progress")),
        (STATUS_PENDING, _("Pending")),
        (STATUS_PAID, _("Paid")),
        (STATUS_PAID_REFUNDED, _("Paid, part refunded")),
        (STATUS_CANCELLED, _("Cancelled")),
        (STATUS_CHARGED_BACK, _("Charged back")),
        (STATUS_REFUNDED, _("Refunded")),
        (STATUS_EXPIRED, _("Expired")),
        (STATUS_UNKNOWN, _("Unknown")),
    )

    merchant_name = models.CharField(_("Docdata account"), max_length=100, default=appsettings.DOCDATA_MERCHANT_NAME)
    merchant_order_id = models.CharField(_("Order ID"), max_length=100, default='')
    order_key = models.CharField(_("Payment cluster ID"), max_length=200, default='', unique=True)

    status = models.CharField(_("Status"), max_length=50, choices=STATUS_CHOICES, default=STATUS_NEW)
    language = models.CharField(_("Language"), max_length=5, blank=True, default='en')

    # Track sent information
    total_gross_amount = models.DecimalField(_("Total gross amount"), max_digits=15, decimal_places=2)
    currency = models.CharField(_("Currency"), max_length=10)
    country = models.CharField(_("Country_code"), max_length=2, null=True, blank=True)

    # Track received information
    total_registered = models.DecimalField(_("Total registered"), max_digits=15, decimal_places=2, default=D('0.00'))
    total_shopper_pending = models.DecimalField(_("Total shopper pending"), max_digits=15, decimal_places=2, default=D('0.00'))
    total_acquirer_pending = models.DecimalField(_("Total acquirer pending"), max_digits=15, decimal_places=2, default=D('0.00'))
    total_acquirer_approved = models.DecimalField(_("Total acquirer approved"), max_digits=15, decimal_places=2, default=D('0.00'))
    total_captured = models.DecimalField(_("Total captured"), max_digits=15, decimal_places=2, default=D('0.00'))
    total_refunded = models.DecimalField(_("Total refunded"), max_digits=15, decimal_places=2, default=D('0.00'))
    total_charged_back = models.DecimalField(_("Total changed back"), max_digits=15, decimal_places=2, default=D('0.00'))

    # Internal info.
    created = models.DateTimeField(_("created"), auto_now_add=True)
    updated = models.DateTimeField(_("updated"), auto_now=True)

    objects = DocdataOrderManager()

    class Meta:
        ordering = ('-created', '-updated')
        verbose_name = _("Docdata Order")
        verbose_name_plural = _("Docdata Orders")

    def __str__(self):
        return self.order_key

    def __repr__(self):
        return "<DocdataOrder: {0}, {1} status={2}>".format(self.order_key, self.merchant_order_id, self.status)

    @property
    def latest_payment(self):
        try:
            return self.payments.order_by('-payment_id').all()[0]
        except IndexError:
            return None

    def cancel(self):
        """
        Cancel an order in Docdata.
        """
        from .facade import get_facade
        facade = get_facade()
        facade.cancel_order(self)

    cancel.alters_data = True


@python_2_unicode_compatible
class DocdataPayment(PolymorphicModel):
    """
    A reported Docdata payment.
    This is a summarized version of a Docdata payment transaction,
    as returned by the status API call.

    Some payment types have additional fields, which are stored as subclass.
    """
    docdata_order = models.ForeignKey(DocdataOrder, related_name='payments')
    payment_id = models.CharField(_("Payment id"), max_length=100, default='', blank=True, primary_key=True)

    # Note: We're not using choices here so that we can write unknown statuses if they are presented by Docdata.
    status = models.CharField(_("status"), max_length=30, default='NEW')

    # The payment method id from Docdata (e.g. IDEAL, MASTERCARD, etc)
    payment_method = models.CharField(max_length=60, default='', blank=True)

    # Track the various amounts associated with this source
    confidence_level = models.CharField(_("Confidence level"), max_length=30, default='', editable=False)
    amount_allocated = models.DecimalField(_("Amount Allocated"), decimal_places=2, max_digits=12, default=D('0.00'), editable=False)
    amount_debited = models.DecimalField(_("Amount Debited"), decimal_places=2, max_digits=12, default=D('0.00'), editable=False)
    amount_refunded = models.DecimalField(_("Amount Refunded"), decimal_places=2, max_digits=12, default=D('0.00'), editable=False)
    amount_chargeback = models.DecimalField(_("Amount Changed back"), decimal_places=2, max_digits=12, default=D('0.00'), editable=False)

    # Internal info.
    created = models.DateTimeField(_("created"), auto_now_add=True)
    updated = models.DateTimeField(_("updated"), auto_now=True)

    def __str__(self):
        return self.payment_id

    class Meta:
        ordering = ('payment_id',)
        verbose_name = _("Payment")
        verbose_name_plural = _("Payments")


# NOTE: currently unused.
# DirectDebit is used for periodic transfers (e.g. "Automatische incasso" in The Netherlands)
class DocdataDirectDebitPayment(DocdataPayment):
    """
    Web direct debit direct payment.
    """
    holder_name = models.CharField(max_length=35)  # max_length from Docdata
    holder_city = models.CharField(max_length=35)  # max_length from Docdata
    holder_country_code = models.CharField(_("Country_code"), max_length=2, null=True, blank=True)

    # Note: there is django-iban for validated versions of these fields.
    # Not needed here.
    iban = models.CharField(max_length=34)
    bic = models.CharField(max_length=11)

    class Meta:
        ordering = ('-created', '-updated')
        verbose_name = _("Direct Debit Payment")
        verbose_name_plural = _("Derect Debit Payments")



import logging
from optparse import make_option
from django.core.management.base import BaseCommand, CommandError

from django.db import transaction
from oscar_docdata.exceptions import DocdataStatusError
from oscar_docdata.facade import get_facade
from oscar_docdata.models import DocdataOrder


class Command(BaseCommand):
    help = "Update the status of the given orders"
    option_list = (
        make_option('--all', action='store_true', dest='all', default=False,
            help="Update all orders"),
        make_option('--status', action='store', dest='status', default=None,
            help="Update all orders of a given status"),
    ) + BaseCommand.option_list


    def handle(self, *args, **options):
        """
        Update the status.
        """
        # At -v2 SOAP requests are outputted.
        do_all = options['all']
        only_status = options['status']
        verbosity = int(options['verbosity'])

        # Apply verbosity
        logging.getLogger('oscar_docdata.interface').setLevel('WARNING' if verbosity < 2 else 'DEBUG')
        logging.getLogger('suds.transport').setLevel('INFO' if verbosity < 3 else 'DEBUG')

        qs = DocdataOrder.objects.active_merchants()
        facade = get_facade()

        if do_all:
            orders = qs.all()

            if only_status:
                orders = orders.filter(status=only_status)

            if args:
                raise CommandError("No order numbers have to be provided for --all")
        else:
            if not args:
                raise CommandError("Expected order numbers as argument")

            # First get all orders, check them.
            orders = []
            for order_number in args:
                try:
                    order = qs.get(merchant_order_id=order_number)
                except DocdataOrder.DoesNotExist:
                    self.stderr.write(u"- Order does not exist: {0}\n".format(order_number))
                    continue
                else:
                    if only_status and order.status != only_status:
                        self.stderr.write(u"- Order {0} does not have status {1}, but {2}\n".format(order_number, only_status, order.status))
                        continue
                    orders.append(order)

        for order in orders:
            self.stdout.write(u"- Checking {0}\n".format(order.merchant_order_id))

            with transaction.atomic():
                # First request the order at docdata, avoid expiring an order which missed an update (very unlikely)
                old_status = order.status
                try:
                    facade.update_order(order)
                except DocdataStatusError as e:
                    self.stderr.write(u"{0}\n".format(e))
                    continue

                if order.status == old_status:
                    self.stderr.write(u"  Order {0} status unchanged, remained: {1}".format(order.merchant_order_id, order.status))
                else:
                    self.stdout.write(u"  Order {0} status changed from {1} to {2}".format(order.merchant_order_id, old_status, order.status))



# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import DataMigration
from django.db import models

class Migration(DataMigration):

    def forwards(self, orm):
        # Fill the merhant name record
        from oscar_docdata import appsettings
        orm['oscar_docdata.docdataorder'].objects.filter(merchant_name='').update(merchant_name=appsettings.DOCDATA_MERCHANT_NAME)

    def backwards(self, orm):
        pass

    models = {
        u'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        u'oscar_docdata.docdatadirectdebitpayment': {
            'Meta': {'ordering': "('-created', '-updated')", 'object_name': 'DocdataDirectDebitPayment', '_ormbases': [u'oscar_docdata.DocdataPayment']},
            'bic': ('django.db.models.fields.CharField', [], {'max_length': '11'}),
            u'docdatapayment_ptr': ('django.db.models.fields.related.OneToOneField', [], {'to': u"orm['oscar_docdata.DocdataPayment']", 'unique': 'True', 'primary_key': 'True'}),
            'holder_city': ('django.db.models.fields.CharField', [], {'max_length': '35'}),
            'holder_country_code': ('django.db.models.fields.CharField', [], {'max_length': '2', 'null': 'True', 'blank': 'True'}),
            'holder_name': ('django.db.models.fields.CharField', [], {'max_length': '35'}),
            'iban': ('django.db.models.fields.CharField', [], {'max_length': '34'})
        },
        u'oscar_docdata.docdataorder': {
            'Meta': {'ordering': "('-created', '-updated')", 'object_name': 'DocdataOrder'},
            'country': ('django.db.models.fields.CharField', [], {'max_length': '2', 'null': 'True', 'blank': 'True'}),
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'currency': ('django.db.models.fields.CharField', [], {'max_length': '10'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'language': ('django.db.models.fields.CharField', [], {'default': "'en'", 'max_length': '5', 'blank': 'True'}),
            'merchant_name': ('django.db.models.fields.CharField', [], {'default': "'solarforsyria_org'", 'max_length': '100'}),
            'merchant_order_id': ('django.db.models.fields.CharField', [], {'default': "''", 'max_length': '100'}),
            'order_key': ('django.db.models.fields.CharField', [], {'default': "''", 'unique': 'True', 'max_length': '200'}),
            'status': ('django.db.models.fields.CharField', [], {'default': "'new'", 'max_length': '50'}),
            'total_acquirer_approved': ('django.db.models.fields.DecimalField', [], {'default': "'0.00'", 'max_digits': '15', 'decimal_places': '2'}),
            'total_acquirer_pending': ('django.db.models.fields.DecimalField', [], {'default': "'0.00'", 'max_digits': '15', 'decimal_places': '2'}),
            'total_captured': ('django.db.models.fields.DecimalField', [], {'default': "'0.00'", 'max_digits': '15', 'decimal_places': '2'}),
            'total_charged_back': ('django.db.models.fields.DecimalField', [], {'default': "'0.00'", 'max_digits': '15', 'decimal_places': '2'}),
            'total_gross_amount': ('django.db.models.fields.DecimalField', [], {'max_digits': '15', 'decimal_places': '2'}),
            'total_refunded': ('django.db.models.fields.DecimalField', [], {'default': "'0.00'", 'max_digits': '15', 'decimal_places': '2'}),
            'total_registered': ('django.db.models.fields.DecimalField', [], {'default': "'0.00'", 'max_digits': '15', 'decimal_places': '2'}),
            'total_shopper_pending': ('django.db.models.fields.DecimalField', [], {'default': "'0.00'", 'max_digits': '15', 'decimal_places': '2'}),
            'updated': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'})
        },
        u'oscar_docdata.docdatapayment': {
            'Meta': {'ordering': "('payment_id',)", 'object_name': 'DocdataPayment'},
            'amount_allocated': ('django.db.models.fields.DecimalField', [], {'default': "'0.00'", 'max_digits': '12', 'decimal_places': '2'}),
            'amount_chargeback': ('django.db.models.fields.DecimalField', [], {'default': "'0.00'", 'max_digits': '12', 'decimal_places': '2'}),
            'amount_debited': ('django.db.models.fields.DecimalField', [], {'default': "'0.00'", 'max_digits': '12', 'decimal_places': '2'}),
            'amount_refunded': ('django.db.models.fields.DecimalField', [], {'default': "'0.00'", 'max_digits': '12', 'decimal_places': '2'}),
            'confidence_level': ('django.db.models.fields.CharField', [], {'default': "''", 'max_length': '30'}),
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'docdata_order': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'payments'", 'to': u"orm['oscar_docdata.DocdataOrder']"}),
            'payment_id': ('django.db.models.fields.CharField', [], {'default': "''", 'max_length': '100', 'primary_key': 'True'}),
            'payment_method': ('django.db.models.fields.CharField', [], {'default': "''", 'max_length': '60', 'blank': 'True'}),
            'polymorphic_ctype': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "u'polymorphic_oscar_docdata.docdatapayment_set'", 'null': 'True', 'to': u"orm['contenttypes.ContentType']"}),
            'status': ('django.db.models.fields.CharField', [], {'default': "'NEW'", 'max_length': '30'}),
            'updated': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'})
        }
    }

    complete_apps = ['oscar_docdata']
    symmetrical = True



from django.contrib import messages
from django.core.urlresolvers import reverse_lazy, reverse
from django.db.models import Q
from django.http import  HttpResponseRedirect
from django.utils.translation import ugettext_lazy as _
from django.views.generic import ListView, DetailView, View, DeleteView
from django.views.generic.detail import SingleObjectMixin
from oscar_docdata.exceptions import DocdataStatusError, DocdataCancelError
from oscar_docdata.facade import get_facade
from oscar_docdata.models import DocdataOrder
from .forms import DocdataOrderSearchForm


class DocdataOrderListView(ListView):
    """
    List of orders
    """
    model = DocdataOrder
    template_name = 'oscar_docdata/dashboard/orders/list.html'
    form_class = DocdataOrderSearchForm
    paginate_by = 25

    def get_queryset(self):
        """
        Build the queryset for this list.
        """
        qs = super(DocdataOrderListView, self).get_queryset()

        # Always show the current merchant only.
        # Can't use the the API calls for different merchants anyway,
        # and this makes it possible to use a multi-tenant site.
        qs = qs.active_merchants()

        # Look for shortcut query filters
        if 'order_number' in self.request.GET or 'order_status' in self.request.GET:
            self.form = self.form_class(self.request.GET)
            if not self.form.is_valid():
                return qs

            data = self.form.cleaned_data
            status = data.get('order_status')
            number = data.get('order_number')

            if number:
                qs = qs.filter(Q(merchant_order_id__contains=number) | Q(order_key__contains=number))
            if status:
                qs = qs.filter(status=status)
        else:
            self.form = self.form_class()

        return qs

    def get_context_data(self, **kwargs):
        ctx = super(DocdataOrderListView, self).get_context_data(**kwargs)
        ctx['form'] = self.form
        return ctx


class DocdataOrderDetailView(DetailView):
    """
    Detail view
    """
    model = DocdataOrder
    template_name = 'oscar_docdata/dashboard/orders/detail.html'
    success_url = reverse_lazy('docdata-order-list')

    def get_context_data(self, **kwargs):
        ctx = super(DocdataOrderDetailView, self).get_context_data(**kwargs)
        ctx['object_payments'] = self.object.payments.all()
        return ctx


class DocdataOrderUpdateStatusView(SingleObjectMixin, View):
    """
    Update the order status
    """
    model = DocdataOrder

    def get(self, request, *args, **kwargs):
        """
        Fetch the latest status from docdata.
        """
        self.object = self.get_object()

        # Perform update.
        try:
            facade = get_facade()
            facade.update_order(self.object)
        except DocdataStatusError as e:
            messages.error(request, e.value)
        else:
            messages.info(request, u"Order status is updated")

        return HttpResponseRedirect(reverse('docdata-order-detail', args=(self.object.pk,)))


class DocdataOrderCancelView(DeleteView):
    """
    Update the order status
    """
    model = DocdataOrder
    template_name_suffix = "_confirm_cancel"
    template_name = 'oscar_docdata/dashboard/orders/confirm_cancel.html'

    def delete(self, request, *args, **kwargs):
        """
        Cancel the object in Docdata
        """
        self.object = self.get_object()

        # Perform cancel
        try:
            facade = get_facade()
            facade.cancel_order(self.object)
        except DocdataCancelError as e:
            messages.error(request, e.value)
        else:
            messages.info(request, _(u'The order "{order_id}" was cancelled successfully.').format(order_id=self.object.merchant_order_id))

        return HttpResponseRedirect(reverse('docdata-order-detail', args=(self.object.pk,)))



"""
Bridging module between Oscar and the gateway module (which is Oscar agnostic)
"""
import logging
from django.utils.translation import get_language
from oscar.apps.order.exceptions import InvalidOrderStatus
from oscar.apps.payment.exceptions import PaymentError
from oscar_docdata import appsettings
from oscar_docdata.compat import get_model
from oscar_docdata.exceptions import DocdataCreateError
from oscar_docdata.gateway import Name, Shopper, Destination, Address, Amount, to_iso639_part1, Invoice, Vat, Item
from oscar_docdata.interface import Interface
from oscar_docdata.utils.load import import_class

__all__ = (
    'get_facade',
    'get_facade_class',
    'Facade',
)

logger = logging.getLogger(__name__)

Order = None
SourceType = None
_FacadeClass = None

def _lazy_get_models():
    # This avoids various import conflicts between apps that may
    # import the Facade before any other models.
    global Order
    global SourceType
    if Order is None:
        Order = get_model('order', 'Order')
        SourceType = get_model('payment', 'SourceType')


def get_facade(*args, **kwargs):
    """
    Get the proper Facade object instance.
    This reads the ``DOCDATA_FACADE_CLASS`` setting.
    :rtype: Facade
    """
    FacadeClass = get_facade_class()
    return FacadeClass(*args, **kwargs)


def get_facade_class():
    """
    Get the proper Facade object instance.
    This reads the ``DOCDATA_FACADE_CLASS`` setting.
    """
    global _FacadeClass
    if _FacadeClass is not None:
        return _FacadeClass

    # Import it.
    _FacadeClass = import_class(appsettings.DOCDATA_FACADE_CLASS, 'DOCDATA_FACADE_CLASS')
    return _FacadeClass



class Facade(Interface):
    """
    The bridge between Oscar and the generic Interface.

    Most methods are just called directly on the Interface.
    """

    def create_payment(self, order_number, total, user, language=None, description=None, profile=None, merchant_name=None, **kwargs):
        """
        Start a new payment session / container.

        :param order_number: The order number generated by Oscar.
        :param total: The price object, including totals and currency.
        :type total: :class:`oscar.core.prices.Price`
        :type user: :class:`django.contrib.auth.models.User`

        Besides the overwritten parameters, also provide:

        :param billing_address: The billing address, used
        :type billing_address: :class:`oscar.apps.address.abstract_models.AbstractAddress`
        :param basket: The basket to submit, to generate the invoice.
        :type basket: :class:`~oscar.apps.basket.abstract_models.AbstractBasket`
        :param shipping_address: The shipping address for the invoice.
        :type shipping_address: :class:`~oscar.apps.address.abstract_models.AbstractAddress`
        :returns: The Docdata order reference ("order key").

        ..note::
            The Docdata API v1.2 has issues with passing the address fields to PayPal.
            It read the "billing_address.state" field as "status", and as hack they
            now pass the "shipping_address.state" to PayPal.
            Hence, the recommended action is to pass the billing address as shipping address too.
        """
        if not profile:
            profile = appsettings.DOCDATA_PROFILE

        try:
            order_key = super(Facade, self).create_payment(order_number, total, user, language=language, description=description, profile=profile, merchant_name=merchant_name, **kwargs)
        except DocdataCreateError as e:
            raise PaymentError(e.value, e)

        return order_key



    def get_create_payment_args(self, order_number, total, user, language=None, description=None, profile=None, **kwargs):
        """
        The arguments for the createpayment call.
        This is a separate method to be easily overwritable.
        """
        if not profile:
            profile = appsettings.DOCDATA_PROFILE

        args = dict(
            order_id=order_number,
            total_gross_amount=Amount(total.incl_tax, total.currency),
            shopper=Shopper(
                id=user.id,
                name=Name(
                    first=user.first_name,
                    last=user.last_name,
                ),
                email=user.email,
                language=to_iso639_part1(language or get_language()),
                gender='U'
            ),
            description=description,
            profile=profile
        )

        # Pass extra fields, needed to handle certain payment types property.
        # That includes PayPal.

        _billingaddress = kwargs.get('billingaddress', None)  # Old name
        billing_address = kwargs.get('billing_address', _billingaddress)
        if billing_address is not None:
            bill_to = Destination.from_address(billing_address)
            # Auto fill in these fields if they are not provided.
            # These fields are not required in Oscar, but Docdata requires them.
            if not bill_to.name.first:
                bill_to.name.first = user.first_name
            if not bill_to.name.last:
                bill_to.name.last = user.last_name

            args['bill_to'] = bill_to

        shipping_address = kwargs.get('shipping_address', None)
        basket = kwargs.get('basket', None)
        if shipping_address is not None and basket is not None:
            invoice = Invoice.from_basket(basket, total, shipping_address)
            # Auto fill in these fields if they are not provided.
            # These fields are not required in Oscar, but Docdata requires them.
            if not invoice.ship_to.name.first:
                invoice.ship_to.name.first = user.first_name
            if not invoice.ship_to.name.last:
                invoice.ship_to.name.last = user.last_name

            args['invoice'] = invoice

        return args


    def order_status_changed(self, docdataorder, old_status, new_status):
        """
        The order status changed.
        """
        _lazy_get_models()
        project_status = appsettings.DOCDATA_ORDER_STATUS_MAPPING.get(new_status, new_status)

        # Update the order in Oscar
        # Using select_for_update() to have a lock on the order first.
        order = Order.objects.select_for_update().get(number=docdataorder.merchant_order_id)
        if order.status == project_status:
            # Parallel update by docdata (return URL and callback), avoid sending the signal twice to the user code.
            logging.info("Order {0} status is already {1}, skipping signal.".format(order.number, order.status))
            return

        old_oscar_status = order.status
        try:
            order.set_status(project_status)
        except InvalidOrderStatus:
            # Avoid forcing a status onto oscar models.
            # A order could be manually updated to a new status (e.g. paid outside of docdata),
            # while the docdata order remains in a pending state.
            logging.warning("Order %s status is %s, will not change to %s", order.number, order.status, project_status)
            return
        else:
            logging.info("Order {0} status changed from {1} to {2}.".format(order.number, old_oscar_status, order.status))

        # Send the signal
        super(Facade, self).order_status_changed(docdataorder, old_status, new_status)


    def get_source_type(self):
        """
        Convenience method, return the canonical SourceType for Docdata payment events.
        """
        _lazy_get_models()
        source_type, _ = SourceType.objects.get_or_create(code='docdata', defaults={'name': "Docdata Payments"})
        return source_type
